# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
from datetime import date, datetime
from typing import Optional, Union

import koji
from specfile.changelog import ChangelogEntry
from specfile.utils import NEVR

from packit.constants import KOJI_BASEURL

logger = logging.getLogger(__name__)


class KojiHelper:
    """
    Class for querying Koji.

    Attributes:
        session: Koji client session.
    """

    def __init__(self, session: Optional[koji.ClientSession] = None) -> None:
        self.session = (
            session if session is not None else koji.ClientSession(baseurl=KOJI_BASEURL)
        )

    def get_builds(self, package: str, since: datetime) -> list[dict]:
        """
        Gets list of builds of a package since the specified datetime.

        Args:
            package: Package name.
            since: Only builds newer than this datetime will be considered.

        Returns:
            List of builds.
        """
        try:
            package_id = self.session.getPackageID(package, strict=True)
        except Exception as e:
            logger.debug(f"Failed to get ID of package {package} from Koji: {e}")
            return []
        try:
            builds = self.session.listBuilds(
                packageID=package_id,
                state=koji.BUILD_STATES["COMPLETE"],
                completeAfter=since.timestamp(),
            )
        except Exception as e:
            logger.debug(f"Failed to get builds of package {package} from Koji: {e}")
            return []
        if not builds:
            logger.debug(f"No builds found for package {package} since {since}")
            return []
        return builds

    def get_nvrs(self, package: str, since: datetime) -> list[str]:
        """
        Gets list of nvr for builds of a package since the specified datetime.

        Args:
            package: Package name.
            since: Only builds newer than this datetime will be considered.

        Returns:
            List of NVRs.
        """
        return [b["nvr"] for b in self.get_builds(package, since)]

    def get_latest_build_in_tag(self, package: str, tag: str) -> Optional[dict]:
        """
        Gets the latest build of a package tagged into the specified tag.

        Args:
            package: Package name.
            tag: Koji tag.

        Returns:
            Latest build or None if there is no such build.
        """
        try:
            builds = self.session.listTagged(
                package=package,
                tag=tag,
                inherit=True,
                latest=True,
                strict=True,
            )
        except Exception as e:
            logger.debug(
                f"Failed to get latest build of package {package} in tag {tag} from Koji: {e}",
            )
            return None
        if not builds:
            return None
        return builds[0]

    def get_latest_nvr_in_tag(self, package: str, tag: str) -> Optional[str]:
        """
        Gets the latest build of a package tagged into the specified tag.

        Args:
            package: Package name.
            tag: Koji tag.

        Returns:
            NVR of the latest build or None if there is no such build.
        """
        build = self.get_latest_build_in_tag(package, tag)
        if not build:
            return None
        return build["nvr"]

    def get_latest_candidate_build(
        self,
        package: str,
        dist_git_branch: str,
    ) -> Optional[dict]:
        """
        Gets the latest build of a package tagged into the candidate tag for the given branch.

        Args:
            package: Package name.
            dist_git_branch: dist-git branch name.

        Returns:
            Latest build or None if there is no such build.
        """
        if not (tag := self.get_candidate_tag(dist_git_branch)):
            return None
        return self.get_latest_build_in_tag(package, tag)

    def get_latest_stable_build(
        self,
        package: str,
        dist_git_branch: str,
        include_candidate: bool = False,
    ) -> Optional[dict]:
        """
        Gets the latest build of a package tagged into any stable or, if requested,
        the candidate tag for the given branch.

        Args:
            package: Package name.
            dist_git_branch: dist-git branch name.
            include_candidate: Whether to consider also builds tagged
              into the corresponding candidate tag.

        Returns:
            Latest build or None if there is no such build.
        """
        if not (candidate_tag := self.get_candidate_tag(dist_git_branch)):
            return None
        tags = self.get_stable_tags(candidate_tag)
        if include_candidate:
            tags.append(candidate_tag)
        return max(
            (self.get_latest_build_in_tag(package, t) for t in tags),
            key=lambda b: NEVR.from_string(b["nvr"]),
        )

    def get_latest_stable_nvr(
        self,
        package: str,
        dist_git_branch: str,
        include_candidate: bool = False,
    ) -> Optional[dict]:
        """
        Gets the NVR of the latest build of a package tagged into any stable or, if requested,
        the candidate tag for the given branch.

        Args:
            package: Package name.
            dist_git_branch: dist-git branch name.
            include_candidate: Whether to consider also builds tagged
              into the corresponding candidate tag.

        Returns:
            NVR of the latest build or None if there is no such build.
        """
        build = self.get_latest_stable_build(
            package,
            dist_git_branch,
            include_candidate,
        )
        if not build:
            return None
        return build["nvr"]

    def get_build_tags(self, nvr: str) -> list[str]:
        """
        Gets tags the specified build is tagged into.

        Args:
            nvr: NVR of the build.

        Returns:
            List of tag names.
        """
        try:
            tags = self.session.listTags(build=nvr)
        except Exception as e:
            logger.debug(f"Failed to get tags of build {nvr} from Koji: {e}")
            return []
        return [t["name"] for t in tags]

    def get_build_changelog(self, nvr: str) -> list[tuple[int, str, str]]:
        """
        Gets changelog associated with SRPM of the specified build.

        Args:
            nvr: NVR of the build.

        Returns:
            List of changelog entries in form of (timestamp, author, content) tuples.
        """
        requested_headers = ["changelogtime", "changelogname", "changelogtext"]
        try:
            headers = self.session.getRPMHeaders(
                rpmID=f"{nvr}.src",
                headers=requested_headers,
                strict=True,
            )
        except Exception as e:
            logger.debug(f"Failed to get changelog of build {nvr} from Koji: {e}")
            return []
        for k, v in headers.items():
            if not isinstance(v, list):
                headers[k] = [v]
        return list(zip(*[headers[h] for h in requested_headers]))

    def get_builds_in_tag(self, tag: str) -> list[dict]:
        """
        Gets list of builds tagged into the specified tag.

        Args:
            tag: Tag name.

        Returns:
            List of builds.
        """
        try:
            builds = self.session.listTagged(
                tag=tag,
                inherit=False,
                latest=False,
                strict=True,
            )
        except Exception as e:
            logger.debug(f"Failed to get builds tagged into {tag} from Koji: {e}")
            return []
        return builds

    def get_build_info(self, build: Union[int, str]) -> Optional[dict]:
        """
        Gets build information.

        Args:
            build: Koji build ID or NVR.

        Returns:
            Build information or None if there is no such build.
        """
        try:
            info = self.session.getBuild(build)
        except Exception as e:
            logger.debug(f"Failed to get build info of {build} from Koji: {e}")
            return None
        return info

    def get_tag_info(self, tag: str) -> Optional[dict]:
        """
        Gets tag information.

        Args:
            tag: Koji tag.

        Returns:
            Tag information or None if there is no such tag.
        """
        try:
            info = self.session.getBuildConfig(tag)
        except Exception as e:
            logger.debug(f"Failed to get tag info of {tag} from Koji: {e}")
            return None
        return info

    def create_sidetag(self, dist_git_branch: str) -> Optional[dict]:
        """
        Creates a new sidetag for the specified dist-git branch.

        Attempts to log in if the underlying session is not authenticated.

        Args:
            dist_git_branch: dist-git branch name.

        Returns:
            New tag information or None if creation failed.
        """
        if not (target := self.get_build_target(dist_git_branch)):
            logger.debug(f"Failed to get build target for {dist_git_branch} from Koji")
            return None
        if not (build_tag := target.get("build_tag_name")):
            logger.debug(f"Failed to get build tag for {dist_git_branch}")
            return None
        if not self.session.logged_in:
            try:
                self.session.gssapi_login()
            except Exception as e:
                logger.debug(f"Authentication failed: {e}")
                return None
        try:
            info = self.session.createSideTag(build_tag)
        except Exception as e:
            logger.debug(f"Failed to create sidetag for {build_tag} in Koji: {e}")
            return None
        return info

    def remove_sidetag(self, sidetag: str) -> None:
        """
        Removes the specified sidetag.

        Attempts to log in if the underlying session is not authenticated.

        Args:
            sidetag: Sidetag name.
        """
        if not self.session.logged_in:
            try:
                self.session.gssapi_login()
            except Exception as e:
                logger.debug(f"Authentication failed: {e}")
                return
        try:
            self.session.removeSideTag(sidetag)
        except Exception as e:
            logger.debug(f"Failed to remove sidetag {sidetag} in Koji: {e}")

    def tag_build(self, nvr: str, tag: str) -> None:
        """
        Tags a build into the specified tag.

        Attempts to log in if the underlying session is not authenticated.

        Args:
            nvr: NVR of the build.
            tag: Tag name.
        """
        if not self.session.logged_in:
            try:
                self.session.gssapi_login()
            except Exception as e:
                logger.debug(f"Authentication failed: {e}")
                return
        try:
            self.session.tagBuild(tag, nvr)
        except Exception as e:
            logger.debug(f"Failed to tag {nvr} into {tag} in Koji: {e}")

    def untag_build(self, nvr: str, tag: str) -> None:
        """
        Untags a build from the specified tag.

        Attempts to log in if the underlying session is not authenticated.

        Args:
            nvr: NVR of the build.
            tag: Tag name.
        """
        if not self.session.logged_in:
            try:
                self.session.gssapi_login()
            except Exception as e:
                logger.debug(f"Authentication failed: {e}")
                return
        try:
            self.session.untagBuild(tag, nvr, strict=True)
        except Exception as e:
            logger.debug(f"Failed to untag {nvr} from {tag} in Koji: {e}")

    def get_build_target(self, dist_git_branch: str) -> Optional[dict]:
        """
        Gets a build target from a dist-git branch name.

        Args:
            dist_git_branch: dist-git branch name.

        Returns:
            Build target or None if not found.
        """
        target_name = self.get_build_target_name(dist_git_branch)
        try:
            target = self.session.getBuildTarget(target_name, strict=True)
        except Exception as e:
            logger.debug(f"Failed to get build target {target_name} from Koji: {e}")
            return None
        return target

    # | Overview of tag names
    # |------------------------------|----------------------|-------------------------|
    # |                              | Fedora               | EPEL                    |
    # |------------------------------|----------------------|-------------------------|
    # | fresh build                  | fN-updates-candidate | epelN-testing-candidate |
    # | associated update in testing | fN-updates-testing   | epelN-testing           |
    # | associated update in stable  | fN-updates/fN        | epelN                   |
    # |------------------------------|----------------------|-------------------------|

    def get_candidate_tag(self, dist_git_branch: str) -> Optional[str]:
        """
        Gets a candidate tag from a dist-git branch name.

        Args:
            dist_git_branch: dist-git branch name.

        Returns:
            Name of matching candidate tag or None if not found.
        """
        if not (target := self.get_build_target(dist_git_branch)):
            logger.debug(f"Failed to get build target for {dist_git_branch} from Koji")
            return None
        if not (dest_tag := target.get("dest_tag_name")):
            logger.debug(f"Failed to get dest tag for {dist_git_branch}")
            return None
        return dest_tag

    def get_stable_tags(self, tag: str) -> list[str]:
        """
        Gets a list of stable tags from the specified tag. Only tags without any suffix
        and tags with "-updates" suffix are considered stable.

        Args:
            tag: Tag name.

        Returns:
            List of stable tags. Can be empty.
        """
        try:
            ancestors = self.session.getFullInheritance(tag)
        except Exception as e:
            logger.debug(f"Failed to get inheritance of {tag} from Koji: {e}")
            return []
        tags = [tag] + [a["name"] for a in ancestors]
        return [t for t in tags if "-" not in t or t.endswith("-updates")]

    @staticmethod
    def get_build_target_name(dist_git_branch: str) -> str:
        """
        Gets a build target name from a dist-git branch name.

        Args:
            dist_git_branch: dist-git branch name.

        Returns:
            Name of matching build target.
        """
        if dist_git_branch in ["rawhide", "main"]:
            return "rawhide"
        return f"{dist_git_branch}-candidate"

    @staticmethod
    def format_changelog(changelog: list[tuple[int, str, str]], since: int = 0) -> str:
        """
        Formats changelog entries since the specified timestamp.

        Args:
            changelog: Changelog as a list of entries in form of
              (timestamp, author, content) tuples.
            since: Only entries newer than this timestamp will be included.

        Returns:
            Formatted changelog as a string.
        """
        lines = []
        for time, name, text in changelog:
            if time <= since:
                break
            timestamp = date.fromtimestamp(time)
            lines.append(
                str(ChangelogEntry.assemble(timestamp, name, text.splitlines())),
            )
        return "\n".join(lines)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import logging
import subprocess
from pathlib import Path
from typing import Optional, Union

import pytest
from distro import linux_distribution
from flexmock import flexmock
from specfile import Specfile
from specfile.changelog import ChangelogEntry

from packit.actions import ActionName
from packit.base_git import PackitRepositoryBase, requests
from packit.command_handler import LocalCommandHandler
from packit.config import CommonPackageConfig, Config, PackageConfig, RunCommandType
from packit.config.sources import SourcesItem
from packit.distgit import DistGit
from packit.local_project import LocalProject, LocalProjectBuilder
from packit.upstream import GitUpstream
from packit.utils import commands
from packit.utils.repo import create_new_repo
from tests.spellbook import can_a_module_be_imported, initiate_git_repo


@pytest.fixture()
def distgit_with_actions():
    return DistGit(
        config=flexmock(Config()),
        package_config=flexmock(
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="beer",
                        actions={
                            ActionName.pre_sync: "command --a",
                            ActionName.get_current_version: "command --b",
                        },
                    ),
                },
            ),
        ),
    )


@pytest.fixture()
def upstream_with_actions():
    return GitUpstream(
        config=flexmock(Config()),
        package_config=flexmock(
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        actions={
                            ActionName.pre_sync: "command --a",
                            ActionName.get_current_version: "command --b",
                        },
                    ),
                },
            ),
        ),
        local_project=flexmock(
            repo_name=flexmock(),
            refresh_the_arguments=lambda: None,
            git_project=flexmock(),
            git_service=flexmock(),
        ),
    )


@pytest.fixture()
def packit_repository_base():
    repo_base = PackitRepositoryBase(
        config=Config(),
        package_config=PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    actions={
                        ActionName.pre_sync: "command --a",
                        ActionName.get_current_version: "command --b",
                    },
                ),
            },
        ),
    )
    repo_base.local_project = flexmock()
    return repo_base


@pytest.fixture()
def packit_repository_base_more_actions():
    return PackitRepositoryBase(
        config=Config(),
        package_config=PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    actions={
                        ActionName.pre_sync: ["command --a", "command --a"],
                        ActionName.get_current_version: "command --b",
                    },
                ),
            },
        ),
    )


@pytest.fixture()
def packit_repository_base_with_sandcastle_object(tmp_path):
    c = Config()
    c.command_handler = RunCommandType.sandcastle
    b = PackitRepositoryBase(
        config=c,
        package_config=PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    actions={
                        ActionName.pre_sync: "command -a",
                        ActionName.get_current_version: "command -b",
                    },
                ),
            },
        ),
    )
    b.local_project = LocalProjectBuilder().build(working_dir=Path("/sandcastle"))
    return b


def test_has_action_upstream(upstream_with_actions):
    assert upstream_with_actions.actions_handler.has_action(ActionName.pre_sync)
    assert not upstream_with_actions.actions_handler.has_action(
        ActionName.create_patches,
    )


def test_has_action_distgit(distgit_with_actions):
    assert distgit_with_actions.actions_handler.has_action(ActionName.pre_sync)
    assert not distgit_with_actions.actions_handler.has_action(
        ActionName.create_patches,
    )


def test_with_action_non_defined(packit_repository_base):
    if packit_repository_base.actions_handler.with_action(
        action=ActionName.create_patches,
    ):
        # this is the style we are using that function
        return

    raise AssertionError()


def test_with_action_defined(packit_repository_base):
    flexmock(commands).should_receive("run_command").once()

    packit_repository_base.local_project = flexmock(working_dir="my/working/dir")

    if packit_repository_base.actions_handler.with_action(action=ActionName.pre_sync):
        # this is the style we are using that function
        raise AssertionError()


def test_with_action_working_dir(packit_repository_base):
    flexmock(LocalCommandHandler).should_receive("run_command").with_args(
        command=["command", "--a"],
        env=None,
        print_live=True,
    ).and_return("command --a").once()

    packit_repository_base.local_project = flexmock(working_dir="my/working/dir")

    assert not packit_repository_base.actions_handler.with_action(
        action=ActionName.pre_sync,
    )


def test_run_action_hook_not_defined(packit_repository_base):
    flexmock(commands).should_receive("run_command").times(0)

    packit_repository_base.local_project = flexmock(working_dir="my/working/dir")

    packit_repository_base.actions_handler.run_action(actions=ActionName.create_patches)


def test_run_action_not_defined(packit_repository_base):
    flexmock(LocalCommandHandler).should_receive("run_command").times(0)

    packit_repository_base.local_project = flexmock(working_dir="my/working/dir")

    action_method = (
        flexmock()
        .should_receive("action_function")
        .with_args("arg", kwarg="kwarg")
        .once()
        .mock()
        .action_function
    )
    packit_repository_base.actions_handler.run_action(
        ActionName.create_patches,
        action_method,
        {},
        "arg",
        kwarg="kwarg",
    )


def test_run_action_defined(packit_repository_base):
    flexmock(LocalCommandHandler).should_receive("run_command").with_args(
        command=["command", "--a"],
        env={},
        print_live=True,
    ).and_return("command --a").once()

    packit_repository_base.local_project = flexmock(working_dir="my/working/dir")

    action_method = (
        flexmock()
        .should_receive("action_function")
        .with_args("arg", kwarg="kwarg")
        .times(0)
        .mock()
        .action_function
    )

    packit_repository_base.actions_handler.run_action(
        ActionName.pre_sync,
        action_method,
        {},
        "arg",
        "kwarg",
    )


@pytest.mark.skipif(
    not can_a_module_be_imported("sandcastle"),
    reason="sandcastle is not installed",
)
def test_run_action_in_sandcastle(
    packit_repository_base_with_sandcastle_object,
    caplog,
):
    from sandcastle import Sandcastle

    flexmock(Sandcastle).should_receive("get_api_client").and_return(None).once()
    flexmock(Sandcastle).should_receive("run").and_return(None).once()

    def mocked_exec(
        command: list[str],
        env: Optional[dict] = None,
        cwd: Optional[Union[str, Path]] = None,
    ):
        if command == ["command", "-b"]:
            return "1.2.3"
        if command == ["command", "-a"]:
            return (
                "make po-pull\n"
                "make[1]: Entering directory "
                "'/sandcastle/docker-io-usercont-sandcastle-prod-20200820-160948197515'\n"
                "TEMP_DIR=$(mktemp --tmpdir -d anaconda-localization-XXXXXXXXXX)\n"
            )
        raise Exception("This command was not expected")

    flexmock(Sandcastle, exec=mocked_exec)
    flexmock(Sandcastle).should_receive("delete_pod").once().and_return(None)
    with caplog.at_level(logging.INFO, logger="packit"):
        packit_repository_base_with_sandcastle_object.actions_handler.run_action(
            ActionName.pre_sync,
            None,
            "arg1",
            "kwarg1",
        )
        packit_repository_base_with_sandcastle_object.actions_handler.run_action(
            ActionName.get_current_version,
            None,
            "arg2",
            "kwarg2",
        )
        # this is being called in PackitAPI.clean
        packit_repository_base_with_sandcastle_object.command_handler.clean()
        # leading space means that we have the output actually printed
        # and it's not a single line with the whole output
        assert " Running command: command -a on dir .\n" in caplog.text
        assert " make po-pull\n" in caplog.text
        assert " anaconda-localization-XXXXXXXXXX)\n" in caplog.text
        assert " 1.2.3\n" in caplog.text


@pytest.mark.skipif(
    not can_a_module_be_imported("sandcastle"),
    reason="sandcastle is not installed",
)
def test_command_handler_is_set(packit_repository_base_with_sandcastle_object):
    from sandcastle import Sandcastle

    flexmock(Sandcastle).should_receive("get_api_client").and_return(None).once()
    flexmock(Sandcastle).should_receive("run").and_return(None).once()

    # it's not set initially
    assert not packit_repository_base_with_sandcastle_object.is_command_handler_set()

    # and should be set once we invoke it
    assert packit_repository_base_with_sandcastle_object.command_handler.sandcastle
    assert packit_repository_base_with_sandcastle_object.is_command_handler_set()


def test_run_action_more_actions(packit_repository_base_more_actions):
    flexmock(LocalCommandHandler).should_receive("run_command").times(2)

    packit_repository_base_more_actions.local_project = flexmock(
        working_dir="my/working/dir",
    )

    action_method = (
        flexmock()
        .should_receive("action_function")
        .with_args("arg", kwarg="kwarg")
        .times(0)
        .mock()
        .action_function
    )
    packit_repository_base_more_actions.actions_handler.run_action(
        ActionName.pre_sync,
        action_method,
        "arg",
        kwarg="kwarg",
    )


def test_get_output_from_action_not_defined(packit_repository_base):
    flexmock(LocalCommandHandler).should_receive("run_command").times(0)

    packit_repository_base.local_project = flexmock(working_dir="my/working/dir")

    result = packit_repository_base.actions_handler.get_output_from_action(
        ActionName.create_patches,
    )
    assert result is None


@pytest.mark.parametrize(
    "source, package_config, expected_urls, extra_source",
    [
        pytest.param(
            "https://download.samba.org/pub/rsync/src/rsync-3.1.3.tar.gz",
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="rsync.spec",
                        sources=[
                            SourcesItem(
                                path="rsync-3.1.3.tar.gz",
                                url="https://git.centos.org/sources/rsync/c8s/82e7829",
                            ),
                        ],
                    ),
                },
                jobs=[],
            ),
            ["https://git.centos.org/sources/rsync/c8s/82e7829"],
            False,
        ),
        pytest.param(
            "https://download.samba.org/pub/rsync/src/rsync-3.1.3.tar.gz",
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="rsync.spec",
                        sources=[
                            SourcesItem(
                                path="rsync-3.1.3.tar.gz",
                                url="https://git.centos.org/sources/rsync/c8s/82e7829",
                            ),
                        ],
                    ),
                },
                jobs=[],
            ),
            [
                "https://git.centos.org/sources/rsync/c8s/82e7829",
                "https://example.com/extra-source.tar.xz",
            ],
            True,
        ),
        pytest.param(
            "rsync-3.1.3.tar.gz",
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="rsync.spec",
                        sources=[
                            SourcesItem(
                                path="rsync-3.1.3.tar.gz",
                                url="https://git.centos.org/sources/rsync/c8s/82e7829",
                            ),
                        ],
                    ),
                },
                jobs=[],
            ),
            ["https://git.centos.org/sources/rsync/c8s/82e7829"],
            False,
        ),
    ],
)
def test_download_remote_sources(
    source,
    package_config,
    expected_urls,
    extra_source,
    tmp_path: Path,
):
    specfile_content = (
        "Name: rsync\n"
        "Version: 3.1.3\n"
        "Release: 1\n"
        f"Source0: {source}\n"
        "%if 0%{?extra_source}\n"
        "Source1: https://example.com/extra-source.tar.xz\n"
        "%endif\n"
        "License: GPLv3+\n"
        "Summary: rsync\n"
        "%description\nrsync\n"
    )
    spec_path = tmp_path / "rsync.spec"
    spec_path.write_text(specfile_content)
    specfile = Specfile(
        spec_path,
        sourcedir=tmp_path,
        autosave=True,
        macros=[("extra_source", "1")] if extra_source else None,
    )
    base_git = PackitRepositoryBase(config=flexmock(), package_config=package_config)
    flexmock(base_git).should_receive("specfile").and_return(specfile)

    expected_urls = iter(expected_urls)
    expected_path = tmp_path / "rsync-3.1.3.tar.gz"

    def mocked_get(url, **_):
        assert url == next(expected_urls)
        return flexmock(
            raise_for_status=lambda: None,
            iter_content=lambda **_: iter([b"1"]),
        )

    flexmock(requests).should_receive("get").replace_with(mocked_get)

    base_git.download_remote_sources()

    flexmock(requests).should_receive("get").and_raise(
        Exception(
            "This should not be called second time since the source is present already.",
        ),
    )
    base_git.download_remote_sources()

    assert expected_path.exists()


def test_set_spec_content(tmp_path):
    distgit_spec_contents = (
        "Name: bring-me-to-the-life\n"
        "Version: 1.0\n"
        "Release: 1\n"
        "Source0: foo.bar\n"
        "License: GPLv3+\n"
        "Summary: evanescence\n"
        "%description\n-\n\n"
        "%changelog\n"
        "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1\n"
        "- Initial package.\n"
    )
    distgit_spec_path = tmp_path / "life.spec"
    distgit_spec_path.write_text(distgit_spec_contents)

    upstream_spec_contents = (
        "Name: bring-me-to-the-life\n"
        "Version: 1.0\n"
        "Release: 1\n"
        "Source0: foo.bor\n"
        "License: MIT\n"
        "Summary: evanescence, I was brought to life\n"
        "%description\n-\n"
        "%changelog\n"
        "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1\n"
        "- Initial package.\n"
    )
    upstream_spec_path = tmp_path / "e-life.spec"
    upstream_spec_path.write_text(upstream_spec_contents)
    upstream_specfile = Specfile(upstream_spec_path, sourcedir=tmp_path, autosave=True)

    dist_git = PackitRepositoryBase(
        config=flexmock(default_parse_time_macros={}),
        package_config=flexmock(
            prerelease_suffix_macro=None,
            prerelease_suffix_pattern=None,
            parse_time_macros={},
        ),
    )
    dist_git._specfile_path = distgit_spec_path

    dist_git.set_specfile_content(upstream_specfile, None, None)
    with dist_git.specfile.sections() as sections:
        assert [
            "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1",
            "- Initial package.",
        ] == sections.changelog
        assert dist_git.specfile.expanded_version == "1.0"
        assert "License: MIT" in sections.package
        assert "Summary: evanescence, I was brought to life" in sections.package

    new_log = ChangelogEntry(
        "* Wed Jun 02 2021 John Fou <john-fou@example.com> - 1.1-1",
        ["- 1.1 upstream release"],
    )
    flexmock(ChangelogEntry).should_receive("assemble").and_return(new_log)
    dist_git.set_specfile_content(upstream_specfile, "1.1", "1.1 upstream release")
    with dist_git.specfile.sections() as sections:
        assert [
            new_log.header,
            new_log.content[0],
            "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1",
            "- Initial package.",
        ] == sections.changelog
    assert dist_git.specfile.expanded_version == "1.1"


@pytest.mark.parametrize(
    "dg_raw_version,dg_raw_release,up_raw_version,up_raw_release,expected_dg_release",
    [
        ("1.0", "2%{?dist}", "1.0", "2%{?dist}", "1"),
        ("1.0", "2%{?dist}", "1.0", "3%{?dist}", "1"),
        ("1.0", "2%{?dist}", "1.1", "2%{?dist}", "1"),
        ("1.0", "2%{?dist}", "1.1", "3%{?dist}", "3"),
        pytest.param(
            "1.0",
            "%autorelease",
            "1.0",
            "%autorelease",
            "%autorelease",
            marks=pytest.mark.skipif(
                linux_distribution()[0].startswith("CentOS"),
                reason="No rpmautospec-rpm-macros installed",
            ),
        ),
        pytest.param(
            "1.0",
            "%{autorelease}",
            "1.1",
            "%{autorelease}",
            "%{autorelease}",
            marks=pytest.mark.skipif(
                linux_distribution()[0].startswith("CentOS"),
                reason="No rpmautospec-rpm-macros installed",
            ),
        ),
        pytest.param(
            "1.0",
            "%autorelease -b 100",
            "1.1",
            "%autorelease -b 100",
            "%autorelease -b 100",
            marks=pytest.mark.skipif(
                linux_distribution()[0].startswith("CentOS"),
                reason="No rpmautospec-rpm-macros installed",
            ),
        ),
        pytest.param(
            "1.0",
            "%autorelease -p -e pre1",
            "1.0",
            "%autorelease",
            "%autorelease",
            marks=pytest.mark.skipif(
                linux_distribution()[0].startswith("CentOS"),
                reason="No rpmautospec-rpm-macros installed",
            ),
        ),
        pytest.param(
            "1.0",
            "2%{?dist}",
            "1.1",
            "%autorelease",
            "1",
            marks=pytest.mark.skipif(
                linux_distribution()[0].startswith("CentOS"),
                reason="No rpmautospec-rpm-macros installed",
            ),
        ),
        pytest.param(
            "1.0",
            "%autorelease",
            "1.1",
            "1%{?dist}",
            "%autorelease",
            marks=pytest.mark.skipif(
                linux_distribution()[0].startswith("CentOS"),
                reason="No rpmautospec-rpm-macros installed",
            ),
        ),
    ],
)
def test_set_spec_content_reset_release(
    tmp_path,
    dg_raw_version,
    dg_raw_release,
    up_raw_version,
    up_raw_release,
    expected_dg_release,
):
    def changelog(release):
        # %autorelease implies %autochangelog
        if "autorelease" in release:
            return "%autochangelog\n"
        return (
            "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1\n"
            "- Initial package.\n"
        )

    distgit_spec_contents = (
        "Name: bring-me-to-the-life\n"
        f"Version: {dg_raw_version}\n"
        f"Release: {dg_raw_release}\n"
        "Source0: foo.bar\n"
        "License: GPLv3+\n"
        "Summary: evanescence\n"
        "%description\n-\n\n"
        "%changelog\n"
    ) + changelog(dg_raw_release)
    distgit_spec_path = tmp_path / "life.spec"
    distgit_spec_path.write_text(distgit_spec_contents)

    upstream_spec_contents = (
        "Name: bring-me-to-the-life\n"
        f"Version: {up_raw_version}\n"
        f"Release: {up_raw_release}\n"
        "Source0: foo.bor\n"
        "License: MIT\n"
        "Summary: evanescence, I was brought to life\n"
        "%description\n-\n"
        "%changelog\n"
    ) + changelog(up_raw_release)
    upstream_spec_path = tmp_path / "e-life.spec"
    upstream_spec_path.write_text(upstream_spec_contents)
    upstream_specfile = Specfile(upstream_spec_path, sourcedir=tmp_path, autosave=True)

    dist_git = PackitRepositoryBase(
        config=flexmock(default_parse_time_macros={}),
        package_config=flexmock(
            prerelease_suffix_macro=None,
            prerelease_suffix_pattern=None,
            parse_time_macros={},
        ),
    )
    dist_git._specfile_path = distgit_spec_path

    dist_git.set_specfile_content(upstream_specfile, "1.1", "1.1 upstream release")
    assert dist_git.specfile.release == expected_dg_release
    if not dist_git.specfile.has_autochangelog:
        with dist_git.specfile.sections() as sections:
            assert f"1.1-{expected_dg_release}" in sections.changelog[0]


@pytest.mark.parametrize("changelog_section", ["\n%changelog\n", ""])
def test_set_spec_content_no_changelog(tmp_path, changelog_section):
    distgit_spec_contents = (
        "Name: bring-me-to-the-life\n"
        "Version: 1.0\n"
        "Release: 1\n"
        "Source0: foo.bar\n"
        "License: GPLv3+\n"
        "Summary: evanescence\n"
        "%description\n-\n" + changelog_section
    )
    distgit_spec_path = tmp_path / "life.spec"
    distgit_spec_path.write_text(distgit_spec_contents)

    upstream_spec_contents = (
        "Name: bring-me-to-the-life\n"
        "Version: 1.0\n"
        "Release: 1\n"
        "Source0: foo.bor\n"
        "License: MIT\n"
        "Summary: evanescence, I was brought to life\n"
        "%description\n-\n"
        "%changelog\n"
        "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1\n"
        "- Initial package.\n"
    )
    upstream_spec_path = tmp_path / "e-life.spec"
    upstream_spec_path.write_text(upstream_spec_contents)
    upstream_specfile = Specfile(upstream_spec_path, sourcedir=tmp_path, autosave=True)

    dist_git = PackitRepositoryBase(
        config=flexmock(default_parse_time_macros={}),
        package_config=flexmock(
            prerelease_suffix_macro=None,
            prerelease_suffix_pattern=None,
            parse_time_macros={},
        ),
    )
    dist_git._specfile_path = distgit_spec_path

    new_log = ChangelogEntry(
        "* Wed Jun 02 2021 John Fou <john-fou@example.com> - 1.1-1",
        ["- 1.1 upstream release"],
    )
    flexmock(ChangelogEntry).should_receive("assemble").and_return(new_log)
    dist_git.set_specfile_content(upstream_specfile, "1.1", "1.1 upstream release")
    with dist_git.specfile.sections() as sections:
        assert sections.changelog == [new_log.header, new_log.content[0]]


@pytest.mark.parametrize(
    "header, raw_version, macro_definitions",
    [
        pytest.param(
            "",
            "1.1",
            {},
        ),
        pytest.param(
            ("%global package_version 1.0\n"),
            "%{package_version}",
            {
                "package_version": "1.1",
            },
        ),
        pytest.param(
            (
                "%global majorver 1\n"
                "%global minorver 0\n"
                "%global package_version %{majorver}.%{minorver}\n"
            ),
            "%{package_version}",
            {
                "majorver": "1",
                "minorver": "1",
                "package_version": "%{majorver}.%{minorver}",
            },
        ),
        pytest.param(
            (
                "%global majorver 1\n"
                "%global minorver 0\n"
                "%global patchver 0\n"
                "%global package_version %{majorver}.%{minorver}.%{patchver}\n"
            ),
            "%{package_version}",
            {
                "majorver": "1",
                "minorver": "0",
                "patchver": "0",
                "package_version": "1.1",
            },
        ),
    ],
)
def test_set_spec_content_version_macros(
    tmp_path,
    header,
    raw_version,
    macro_definitions,
):
    distgit_spec_contents = header + (
        "Name: bring-me-to-the-life\n"
        "Version: %{package_version}\n"
        "Release: 1\n"
        "Source0: foo.bar\n"
        "License: GPLv3+\n"
        "Summary: evanescence\n"
        "%description\n-\n\n"
        "%changelog\n"
        "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1\n"
        "- Initial package.\n"
    )
    distgit_spec_path = tmp_path / "life.spec"
    distgit_spec_path.write_text(distgit_spec_contents)

    upstream_spec_contents = header + (
        "Name: bring-me-to-the-life\n"
        "Version: %{package_version}\n"
        "Release: 1\n"
        "Source0: foo.bor\n"
        "License: MIT\n"
        "Summary: evanescence, I was brought to life\n"
        "%description\n-\n"
        "%changelog\n"
        "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1\n"
        "- Initial package.\n"
    )
    upstream_spec_path = tmp_path / "e-life.spec"
    upstream_spec_path.write_text(upstream_spec_contents)
    upstream_specfile = Specfile(upstream_spec_path, sourcedir=tmp_path, autosave=True)

    dist_git = PackitRepositoryBase(
        config=flexmock(default_parse_time_macros={}),
        package_config=flexmock(
            prerelease_suffix_macro=None,
            prerelease_suffix_pattern=None,
            parse_time_macros={},
        ),
    )
    dist_git._specfile_path = distgit_spec_path

    new_log = ChangelogEntry(
        "* Wed Jun 02 2021 John Fou <john-fou@example.com> - 1.1-1",
        ["- 1.1 upstream release"],
    )
    flexmock(ChangelogEntry).should_receive("assemble").and_return(new_log)
    dist_git.set_specfile_content(upstream_specfile, "1.1", "1.1 upstream release")
    with dist_git.specfile.sections() as sections:
        assert [
            new_log.header,
            new_log.content[0],
            "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1",
            "- Initial package.",
        ] == sections.changelog
    assert dist_git.specfile.expanded_version == "1.1"
    assert dist_git.specfile.version == raw_version
    with dist_git.specfile.macro_definitions() as mds:
        for md in mds:
            assert macro_definitions.pop(md.name) == md.body
    assert not macro_definitions


@pytest.mark.parametrize(
    "header, version, source_url, prerelease_suffix_pattern, prerelease_suffix_macro, "
    "upstream_version, spec_version, is_prerelease",
    [
        pytest.param(
            "%define uversion %{version_no_tilde %{quote:%nil}}\n",
            "1.2.3~a4",
            "https://example.com/files/v%{uversion}/%{name}-%{uversion}.tar.gz",
            r"()a\d+",
            None,
            "1.2.3",
            "1.2.3",
            False,
        ),
        pytest.param(
            "%define uversion %{version_no_tilde %{quote:%nil}}\n",
            "1.2.3",
            "https://example.com/files/v%{uversion}/%{name}-%{uversion}.tar.gz",
            r"()a\d+",
            None,
            "1.2.4a1",
            "1.2.4~a1",
            True,
        ),
        pytest.param(
            "",
            "1.2.3~b4",
            "https://example.com/files/%{name}-%{version_no_tilde}.tar.gz",
            r"(-)b\d+",
            None,
            "1.2.3",
            "1.2.3",
            False,
        ),
        pytest.param(
            "",
            "1.2.3",
            "https://example.com/files/%{name}-%{version_no_tilde}.tar.gz",
            r"(-)b\d+",
            None,
            "1.2.4-b1",
            "1.2.4~b1",
            True,
        ),
        pytest.param(
            (
                "%global base_version 1.2.3\n"
                "%global prerelease rc4\n"
                "%global package_version %{base_version}%{?prerelease:~%{prerelease}}\n"
                "%global tarball_version %{base_version}%{?prerelease}\n"
            ),
            "%{package_version}",
            "https://example.com/files/v%{base_version}/%{name}-%{tarball_version}.tar.gz",
            r"()rc\d+",
            "prerelease",
            "1.2.3",
            "1.2.3",
            False,
        ),
        pytest.param(
            (
                "%global base_version 1.2.3\n"
                "#global prerelease rc4\n"
                "%global package_version %{base_version}%{?prerelease:~%{prerelease}}\n"
                "%global tarball_version %{base_version}%{?prerelease}\n"
            ),
            "%{package_version}",
            "https://example.com/files/v%{base_version}/%{name}-%{tarball_version}.tar.gz",
            r"()rc\d+",
            "prerelease",
            "1.2.4rc1",
            "1.2.4~rc1",
            True,
        ),
        pytest.param(
            (
                "%global majorver 1\n"
                "%global minorver 2\n"
                "%global patchver 3\n"
                "%global prerel beta\n"
                "%global package_version "
                "%{majorver}.%{minorver}.%{patchver}%{?prerelease:~%{prerelease}}\n"
                "%global tarball_version "
                "%{majorver}.%{minorver}.%{patchver}%{?prerelease:-%{prerelease}}\n"
            ),
            "%{package_version}",
            "https://example.com/files/v%{majorver}.%{minorver}/%{name}-%{tarball_version}.tar.gz",
            r"(-)beta\d*",
            "prerel",
            "1.2.3",
            "1.2.3",
            False,
        ),
        pytest.param(
            (
                "%global majorver 1\n"
                "%global minorver 2\n"
                "%global patchver 3\n"
                "%global prerel beta\n"
                "%global package_version %{majorver}.%{minorver}.%{patchver}%{?prerel:~%{prerel}}\n"
                "%global tarball_version %{majorver}.%{minorver}.%{patchver}%{?prerel:-%{prerel}}\n"
            ),
            "%{package_version}",
            "https://example.com/files/v%{majorver}.%{minorver}/%{name}-%{tarball_version}.tar.gz",
            r"(-)beta\d*",
            "prerel",
            "1.2.4-beta",
            "1.2.4~beta",
            True,
        ),
        pytest.param(
            (
                "%global basever 1.2.3\n"
                "%global prerel alpha_4\n"
                "%if 0%{?prerel:1}\n"
                "%global package_version %{basever}~%{prerel}\n"
                "%global tarball_version %{basever}.%{prerel}\n"
                "%else\n"
                "%global package_version %{basever}\n"
                "%global tarball_version %{basever}\n"
                "%endif\n"
            ),
            "%{package_version}",
            "https://example.com/files/%{name}-%{tarball_version}.tar.gz",
            r"(\.)alpha_\d+",
            "prerel",
            "1.2.3",
            "1.2.3",
            False,
        ),
        pytest.param(
            (
                "%global basever 1.2.3\n"
                "%global prerel alpha_4\n"
                "%if 0%{?prerel:1}\n"
                "%global package_version %{basever}~%{prerel}\n"
                "%global tarball_version %{basever}.%{prerel}\n"
                "%else\n"
                "%global package_version %{basever}\n"
                "%global tarball_version %{basever}\n"
                "%endif\n"
            ),
            "%{package_version}",
            "https://example.com/files/%{name}-%{tarball_version}.tar.gz",
            r"(\.)alpha_\d+",
            "prerel",
            "1.2.4.alpha_1",
            "1.2.4~alpha_1",
            True,
        ),
    ],
)
def test_set_spec_content_prerelease(
    tmp_path,
    header,
    version,
    source_url,
    prerelease_suffix_pattern,
    prerelease_suffix_macro,
    upstream_version,
    spec_version,
    is_prerelease,
):
    distgit_spec_contents = header + (
        "Name: bring-me-to-the-life\n"
        f"Version: {version}\n"
        "Release: 1\n"
        f"Source0: {source_url}\n"
        "License: GPLv3+\n"
        "Summary: evanescence\n"
        "%description\n-\n\n"
        "%changelog\n"
        "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1\n"
        "- Initial package.\n"
    )
    distgit_spec_path = tmp_path / "life.spec"
    distgit_spec_path.write_text(distgit_spec_contents)

    upstream_spec_contents = header + (
        "Name: bring-me-to-the-life\n"
        f"Version: {version}\n"
        "Release: 1\n"
        f"Source0: {source_url}\n"
        "License: MIT\n"
        "Summary: evanescence, I was brought to life\n"
        "%description\n-\n"
        "%changelog\n"
        "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1\n"
        "- Initial package.\n"
    )
    upstream_spec_path = tmp_path / "e-life.spec"
    upstream_spec_path.write_text(upstream_spec_contents)
    upstream_specfile = Specfile(upstream_spec_path, sourcedir=tmp_path, autosave=True)

    dist_git = PackitRepositoryBase(
        config=flexmock(default_parse_time_macros={}),
        package_config=flexmock(
            prerelease_suffix_macro=prerelease_suffix_macro,
            prerelease_suffix_pattern=prerelease_suffix_pattern,
            parse_time_macros={},
        ),
    )
    dist_git._specfile_path = distgit_spec_path

    new_log = ChangelogEntry(
        f"* Wed Jun 02 2021 John Fou <john-fou@example.com> - {spec_version}-1",
        [f"- {upstream_version} upstream release"],
    )
    flexmock(ChangelogEntry).should_receive("assemble").and_return(new_log)
    dist_git.set_specfile_content(
        upstream_specfile,
        upstream_version,
        f"{upstream_version} upstream release",
    )
    with dist_git.specfile.sections() as sections:
        assert [
            new_log.header,
            new_log.content[0],
            "* Mon Mar 04 2019 Foo Bor <foo-bor@example.com> - 1.0-1",
            "- Initial package.",
        ] == sections.changelog
    assert dist_git.specfile.expanded_version == spec_version
    with dist_git.specfile.sources() as sources:
        assert sources[0].location == source_url
        assert (
            sources[0].expanded_filename
            == f"bring-me-to-the-life-{upstream_version}.tar.gz"
        )
    if prerelease_suffix_macro:
        with dist_git.specfile.macro_definitions() as mds:
            assert mds.get(prerelease_suffix_macro).commented_out == (not is_prerelease)


@pytest.mark.parametrize(
    "remote_branches,branch_found",
    [
        pytest.param([], False, id="No remote branches already opened"),
        pytest.param(
            ["one", "two", "three"],
            False,
            id="Remote branches not from packit",
        ),
        pytest.param(
            ["rawhide-pull_from_upstream"],
            True,
            id="New packit remote branch matching",
        ),
    ],
)
def test_search_branch(tmp_path, remote_branches, branch_found):
    remote = tmp_path / "remote_repo"
    remote.mkdir()
    create_new_repo(remote, [])

    for branch in remote_branches:
        subprocess.check_call(["git", "checkout", "-b", branch], cwd=remote)
        subprocess.check_call(["touch", branch], cwd=remote)
        subprocess.check_call(["git", "add", branch], cwd=remote)
        subprocess.check_call(["git", "commit", "-m", branch], cwd=remote)

    local = tmp_path / "local_repo"
    initiate_git_repo(local, remotes=[("origin", remote / ".git")])

    dist_git = PackitRepositoryBase(
        config=flexmock(),
        package_config=flexmock(),
    )
    dist_git.local_project = LocalProject(working_dir=local)
    assert dist_git.search_branch("rawhide-pull_from_upstream") == branch_found


@pytest.mark.parametrize(
    "remote_branches",
    [
        pytest.param([], id="No remote branches already opened"),
        pytest.param(
            ["rawhide-pull_from_upstream"],
            id="Packit remote branch already exist",
        ),
    ],
)
def test_checkout_branch(tmp_path, remote_branches):
    remote = tmp_path / "remote_repo"
    remote.mkdir()
    create_new_repo(remote, [])

    for branch in remote_branches:
        subprocess.check_call(["git", "checkout", "-b", branch], cwd=remote)
        subprocess.check_call(["touch", branch], cwd=remote)
        subprocess.check_call(["git", "add", branch], cwd=remote)
        subprocess.check_call(["git", "commit", "-m", branch], cwd=remote)

    local = tmp_path / "local_repo"
    initiate_git_repo(local, remotes=[("origin", remote / ".git")])

    dist_git = PackitRepositoryBase(
        config=flexmock(),
        package_config=flexmock(),
    )
    dist_git.local_project = LocalProject(working_dir=local)
    dist_git.checkout_branch("rawhide-pull_from_upstream")


@pytest.mark.parametrize(
    "package_config_macros,default_macros,result",
    [
        pytest.param({}, {"fedora": "0"}, [("fedora", "0")]),
        pytest.param({"fedora": "39"}, {"fedora": "0"}, [("fedora", "39")]),
    ],
)
def test_default_macro_definitions(
    tmp_path,
    package_config_macros,
    default_macros,
    result,
):
    distgit_spec_contents = (
        "Name: bring-me-to-the-life\n"
        "Version: 1\n"
        "Release: 1\n"
        "Source0: foo.bar\n"
        "License: GPLv3+\n"
        "Summary: evanescence\n"
        "%description\n-\n\n"
        "%changelog\n"
    )
    distgit_spec_path = tmp_path / "life.spec"
    distgit_spec_path.write_text(distgit_spec_contents)

    dist_git = PackitRepositoryBase(
        config=flexmock(default_parse_time_macros=default_macros),
        package_config=flexmock(
            prerelease_suffix_macro=None,
            prerelease_suffix_pattern=None,
            parse_time_macros=package_config_macros,
        ),
    )
    dist_git._specfile_path = distgit_spec_path
    assert dist_git.specfile.macros == result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.cli.dist_git_init import DistGitInitializer


@pytest.mark.parametrize(
    "initializer,expected_config_dict",
    [
        pytest.param(
            DistGitInitializer(
                config=None,
                path_or_url=None,
                upstream_git_url="my-url",
            ),
            {
                "upstream_project_url": "my-url",
                "jobs": [
                    {
                        "job": "pull_from_upstream",
                        "trigger": "release",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "koji_build",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "bodhi_update",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                ],
            },
            id="only-url",
        ),
        pytest.param(
            DistGitInitializer(
                config=None,
                path_or_url=None,
                upstream_git_url="my-url",
                upstream_tag_exclude="regex-1",
                upstream_tag_include="regex-2",
                upstream_tag_template="v{version}",
                issue_repository="issue-repo-url",
            ),
            {
                "upstream_project_url": "my-url",
                "upstream_tag_template": "v{version}",
                "upstream_tag_include": "regex-2",
                "upstream_tag_exclude": "regex-1",
                "issue_repository": "issue-repo-url",
                "jobs": [
                    {
                        "job": "pull_from_upstream",
                        "trigger": "release",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "koji_build",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "bodhi_update",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                ],
            },
            id="basic-options",
        ),
        pytest.param(
            DistGitInitializer(
                config=None,
                path_or_url=None,
                upstream_git_url="my-url",
                dist_git_branches="fedora-38,fedora-39",
            ),
            {
                "upstream_project_url": "my-url",
                "jobs": [
                    {
                        "job": "pull_from_upstream",
                        "trigger": "release",
                        "dist_git_branches": ["fedora-38", "fedora-39"],
                    },
                    {
                        "job": "koji_build",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-38", "fedora-39"],
                    },
                    {
                        "job": "bodhi_update",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-38", "fedora-39"],
                    },
                ],
            },
            id="dist-git-branches",
        ),
        pytest.param(
            DistGitInitializer(
                config=None,
                path_or_url=None,
                upstream_git_url="my-url",
                allowed_committers="admin1,admin2",
            ),
            {
                "upstream_project_url": "my-url",
                "allowed_committers": ["admin1", "admin2"],
                "jobs": [
                    {
                        "job": "pull_from_upstream",
                        "trigger": "release",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "koji_build",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "bodhi_update",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                ],
            },
            id="allowed-committers",
        ),
        pytest.param(
            DistGitInitializer(
                config=None,
                path_or_url=None,
                upstream_git_url="my-url",
                allowed_pr_authors="admin1,admin2",
            ),
            {
                "upstream_project_url": "my-url",
                "allowed_pr_authors": ["admin1", "admin2"],
                "jobs": [
                    {
                        "job": "pull_from_upstream",
                        "trigger": "release",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "koji_build",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "bodhi_update",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                ],
            },
            id="allowed_pr_authors",
        ),
        pytest.param(
            DistGitInitializer(
                config=None,
                path_or_url=None,
                upstream_git_url="my-url",
                no_bodhi_update=True,
            ),
            {
                "upstream_project_url": "my-url",
                "jobs": [
                    {
                        "job": "pull_from_upstream",
                        "trigger": "release",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                    {
                        "job": "koji_build",
                        "trigger": "commit",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                ],
            },
            id="no-bodhi",
        ),
        pytest.param(
            DistGitInitializer(
                config=None,
                path_or_url=None,
                upstream_git_url="my-url",
                no_koji_build=True,
            ),
            {
                "upstream_project_url": "my-url",
                "jobs": [
                    {
                        "job": "pull_from_upstream",
                        "trigger": "release",
                        "dist_git_branches": ["fedora-rawhide"],
                    },
                ],
            },
            id="no-koji",
        ),
    ],
)
def test_generate_pacakge_config_dict(
    initializer: DistGitInitializer,
    expected_config_dict: dict,
):
    assert initializer.generate_package_config_dict() == expected_config_dict


def test_parse_actions_from_file(tmp_path):
    actions_file = tmp_path / "actions.yaml"
    actions_file.write_text(
        """\
changelog-entry:
- bash -c 'some command'
""",
    )
    initializer = DistGitInitializer(
        config=None,
        path_or_url=None,
        upstream_git_url=None,
        actions_file=actions_file,
    )
    assert initializer.parse_actions_from_file() == {
        "changelog-entry": ["bash -c 'some command'"],
    }



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from bodhi.client.bindings import BodhiClient
from flexmock import flexmock

from packit.status import Status


def test_status_updates(config_mock, package_config_mock, upstream_mock, distgit_mock):
    flexmock(
        BodhiClient,
        query=lambda packages, page: {
            "updates": [
                {
                    "title": "python-requre-0.8.1-2.fc33",
                    "karma": 2,
                    "status": "stable",
                    "release": {"branch": "f33"},
                },
                {
                    "title": "python-requre-0.8.1-2.fc34",
                    "karma": 3,
                    "status": "stable",
                    "release": {"branch": "f34"},
                },
            ],
            "page": 1,
            "pages": 1,
        },
    )

    status = Status(config_mock, package_config_mock, upstream_mock, distgit_mock)
    table = status.get_updates()
    assert table == [
        ["python-requre-0.8.1-2.fc33", 2, "stable"],
        ["python-requre-0.8.1-2.fc34", 3, "stable"],
    ]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy
from pathlib import Path, PosixPath
from typing import Optional

import pytest
from flexmock import flexmock
from github import GithubException
from marshmallow import ValidationError
from ogr.abstract import GitProject, GitService
from ogr.exceptions import GithubAPIException

import packit.config.package_config
from packit.actions import ActionName
from packit.config import (
    CommonPackageConfig,
    JobConfig,
    JobConfigTriggerType,
    JobType,
    get_package_config_from_repo,
)
from packit.config.package_config import (
    PackageConfig,
    find_remote_package_config,
    get_local_package_config,
    get_local_specfile_path,
    get_specfile_path_from_repo,
)
from packit.config.sources import SourcesItem
from packit.constants import CONFIG_FILE_NAMES
from packit.exceptions import PackitConfigException
from packit.schema import PackageConfigSchema
from packit.sync import SyncFilesItem
from tests.spellbook import SYNC_FILES, UP_OSBUILD
from tests.unit.config.test_config import (
    get_default_job_config,
    get_job_config_build_for_branch,
    get_job_config_dict_build_for_branch,
    get_job_config_dict_full,
    get_job_config_dict_simple,
    get_job_config_full,
    get_job_config_simple,
)


@pytest.fixture()
def job_config_simple():
    return get_job_config_simple()


@pytest.mark.parametrize(
    "files,expected",
    [(["foo.spec"], "foo.spec"), ([], None)],
)
def test_get_specfile_path_from_repo(files, expected):
    gp = flexmock(GitProject)
    gp.should_receive("full_repo_name").and_return("a/b")
    gp.should_receive("get_files").and_return(files)
    git_project = GitProject(repo="", service=GitService(), namespace="")
    assert get_specfile_path_from_repo(project=git_project) == expected


@pytest.mark.parametrize(
    "package_config, project",
    [
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                            ),
                        },
                    ),
                ],
            ),
            None,
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                project="example",
                            ),
                        },
                    ),
                ],
            ),
            "example",
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.release,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                project="example1",
                            ),
                        },
                    ),
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                project="example2",
                            ),
                        },
                    ),
                ],
            ),
            "example1",
        ),
    ],
)
def test_project_from_copr_build_job(package_config, project):
    config_project_value = package_config.get_copr_build_project_value()
    assert config_project_value == project


@pytest.mark.parametrize(
    "package_config, expected",
    [
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                dist_git_branches=[
                                    "example",
                                ],
                            ),
                        },
                    ),
                ],
            ),
            [
                "example",
            ],
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                                dist_git_branches=[
                                    "example1",
                                    "example2",
                                ],
                            ),
                        },
                    ),
                ],
            ),
            ["example1", "example2"],
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="xxx",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.propose_downstream,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                specfile_path="xxx",
                            ),
                        },
                    ),
                ],
            ),
            [],
        ),
    ],
)
def test_dg_branches_from_propose_downstream_job(package_config, expected):
    branches = package_config.get_propose_downstream_dg_branches_value()
    assert branches == expected


def test_package_config_equal(job_config_simple):
    assert PackageConfig(
        packages={
            "package": CommonPackageConfig(
                specfile_path="fedora/package.spec",
                synced_files=[SyncFilesItem(src=["packit.yaml"], dest="packit.yaml")],
                downstream_package_name="package",
                create_pr=True,
            ),
        },
        jobs=[job_config_simple],
    ) == PackageConfig(
        packages={
            "package": CommonPackageConfig(
                specfile_path="fedora/package.spec",
                synced_files=[SyncFilesItem(src=["packit.yaml"], dest="packit.yaml")],
                downstream_package_name="package",
                create_pr=True,
            ),
        },
        jobs=[job_config_simple],
    )


@pytest.mark.parametrize(
    "not_equal_package_config",
    [
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/other-package.spec",
                    synced_files=[
                        SyncFilesItem(src=["a"], dest="a"),
                        SyncFilesItem(src=["b"], dest="b"),
                    ],
                ),
            },
            jobs=[get_job_config_simple()],
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    synced_files=[SyncFilesItem(src=["c"], dest="c")],
                ),
            },
            jobs=[get_job_config_simple()],
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    synced_files=[
                        SyncFilesItem(src=["a"], dest="a"),
                        SyncFilesItem(src=["b"], dest="b"),
                    ],
                ),
            },
            jobs=[get_job_config_full()],
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    synced_files=[
                        SyncFilesItem(src=["c"], dest="c"),
                        SyncFilesItem(src=["d"], dest="d"),
                    ],
                    create_pr=False,
                ),
            },
            jobs=[get_job_config_full()],
        ),
    ],
)
def test_package_config_not_equal(not_equal_package_config):
    config = PackageConfig(
        packages={
            "package": CommonPackageConfig(
                specfile_path="fedora/package.spec",
                synced_files=[
                    SyncFilesItem(src=["c"], dest="c"),
                    SyncFilesItem(src=["d"], dest="d"),
                ],
                create_pr=True,
            ),
        },
        jobs=[get_job_config_full()],
    )
    assert config != not_equal_package_config


@pytest.mark.parametrize(
    "raw,is_valid",
    [
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "synced_files": "fedora/foobar.spec",
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/foobar.spec"],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/foobar.spec", "somefile", "somedirectory"],
                "jobs": [],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "dist_git_branches": [
                            "fedora-all",
                        ],
                    },
                ],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "dist_git_branches": ["fedora-all", "epel-8"],
                    },
                ],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "targets": [
                            "fedora-stable",
                        ],
                    },
                ],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "targets": ["fedora-stable", "fedora-development"],
                    },
                ],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "targets": ["f31", "f32"],
                        "timeout": 123,
                        "owner": "santa",
                        "project": "gifts",
                    },
                ],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "tests",
                        "trigger": "pull_request",
                        "targets": [
                            "fedora-all",
                        ],
                        "env": {
                            "MYVAR1": 5,
                            "MYVAR2": "foo",
                        },
                    },
                ],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/foobar.spec"],
                "actions": {
                    "pre-sync": "some/pre-sync/command --option",
                    "get-current-version": "get-me-version",
                },
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/foobar.spec"],
                "actions": {
                    "pre-sync": "some/pre-sync/command --option",
                    "unknown-action": "nothing",
                },
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "actions": ["actions" "has", "to", "be", "key", "value"],
                "jobs": [{"job": "asd", "trigger": "qwe"}],
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "notifications": {"pull_request": {"successful_build": False}},
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "notifications": {
                    "pull_request": {"successful_build": False},
                    "failure_comment": {"message": "my comment"},
                },
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "notifications": {"failure_comment": {"message": "my comment"}},
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "notifications": {"pull_request": {"successful_build": "nie"}},
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "notifications": {"pull_request": False},
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "sources": [{"path": "example_path"}],
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "sources": [{"url": "example_url"}],
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "sources": {"path": "example_path", "url": "example_url"},
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "targets": [
                            "fedora-stable",
                        ],
                        "module_hotfixes": True,
                    },
                ],
            },
            True,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "targets": [
                            "fedora-stable",
                        ],
                        "module_hotfixes": True,
                        "status_name_template": "packit-copr-build-release",
                    },
                ],
            },
            True,
        ),
    ],
)
def test_package_config_validate(raw, is_valid):
    if not is_valid:
        with pytest.raises((ValidationError, ValueError)):
            PackageConfig.get_from_dict(raw)
    else:
        PackageConfig.get_from_dict(raw)


@pytest.mark.parametrize(
    "raw,is_valid",
    [
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "unknown": "key",
                    },
                ],
            },
            False,
        ),
        (
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "unknown": "key",
                    },
                ],
            },
            False,
        ),
    ],
)
def test_package_config_validate_unknown_key(raw, is_valid):
    if not is_valid:
        with pytest.raises((ValidationError, ValueError)):
            PackageConfig.get_from_dict(raw)
    else:
        PackageConfig.get_from_dict(raw)


@pytest.mark.parametrize(
    "raw",
    [
        # {"specfile_path": "test/spec/file/path", "something": "different"},
        {
            "specfile_path": "test/spec/file/path",
            "jobs": [{"trigger": "release", "release_to": ["f28"]}],
        },
    ],
)
def test_package_config_parse_error(raw):
    with pytest.raises(Exception):
        PackageConfig.get_from_dict(raw_dict=raw)


@pytest.mark.parametrize(
    "raw,expected",
    [
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "jobs": [get_job_config_dict_full()],
                "downstream_package_name": "package",
                "create_pr": False,
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="fedora/package.spec",
                        downstream_package_name="package",
                        create_pr=False,
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                        ],
                    ),
                },
                jobs=[
                    get_job_config_full(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        create_pr=False,
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                        ],
                    ),
                ],
            ),
            id="specfile_path+synced_files+job_config_full+downstream_package_name+create_pr",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": [
                    "fedora/package.spec",
                    "somefile",
                    "other",
                    "directory/files",
                ],
                "jobs": [get_job_config_dict_simple()],
                "downstream_package_name": "package",
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                            SyncFilesItem(src=["somefile"], dest="somefile"),
                            SyncFilesItem(src=["other"], dest="other"),
                            SyncFilesItem(
                                src=["directory/files"],
                                dest="directory/files",
                            ),
                        ],
                    ),
                },
                jobs=[
                    get_job_config_simple(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                            SyncFilesItem(src=["somefile"], dest="somefile"),
                            SyncFilesItem(src=["other"], dest="other"),
                            SyncFilesItem(
                                src=["directory/files"],
                                dest="directory/files",
                            ),
                        ],
                    ),
                ],
            ),
            id="specfile_path+synced_files+job_config_dict_simple+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "jobs": [get_job_config_dict_full()],
                "downstream_package_name": "package",
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                        ],
                    ),
                },
                jobs=[
                    get_job_config_full(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                        ],
                    ),
                ],
            ),
            id="specfile_path+synced_files(spec_only)+job_config_full+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec", "somefile"],
                "jobs": [get_job_config_dict_full()],
                "downstream_package_name": "package",
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                            SyncFilesItem(src=["somefile"], dest="somefile"),
                        ],
                    ),
                },
                jobs=[
                    get_job_config_full(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                            SyncFilesItem(src=["somefile"], dest="somefile"),
                        ],
                    ),
                ],
            ),
            id="specfile_path+synced_files+job_config_full+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "jobs": [get_job_config_dict_full()],
                "upstream_project_url": "https://github.com/asd/qwe",
                "upstream_package_name": "qwe",
                "dist_git_base_url": "https://something.wicked",
                "downstream_package_name": "package",
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        upstream_project_url="https://github.com/asd/qwe",
                        upstream_package_name="qwe",
                        dist_git_base_url="https://something.wicked",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                        ],
                    ),
                },
                jobs=[
                    get_job_config_full(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        upstream_project_url="https://github.com/asd/qwe",
                        upstream_package_name="qwe",
                        dist_git_base_url="https://something.wicked",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                        ],
                    ),
                ],
            ),
            id="specfile_path+synced_files+job_config_dict_full+upstream_project_url"
            "+upstream_package_name+dist_git_base_url+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "actions": {
                    "pre-sync": "some/pre-sync/command --option",
                    "get-current-version": "get-me-version",
                },
                "jobs": [],
                "upstream_project_url": "https://github.com/asd/qwe",
                "upstream_package_name": "qwe",
                "dist_git_base_url": "https://something.wicked",
                "downstream_package_name": "package",
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="fedora/package.spec",
                        actions={
                            ActionName.pre_sync: "some/pre-sync/command --option",
                            ActionName.get_current_version: "get-me-version",
                        },
                        upstream_project_url="https://github.com/asd/qwe",
                        upstream_package_name="qwe",
                        dist_git_base_url="https://something.wicked",
                        downstream_package_name="package",
                    ),
                },
                jobs=[],
            ),
            id="specfile_path+actions+empty_jobs+upstream_project_url"
            "+upstream_package_name+dist_git_base_url+downstream_package_name",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "downstream_package_name": "package",
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                        ],
                    ),
                },
                jobs=get_default_job_config(
                    downstream_package_name="package",
                    specfile_path="fedora/package.spec",
                    synced_files=[
                        SyncFilesItem(
                            src=["fedora/package.spec"],
                            dest="fedora/package.spec",
                        ),
                    ],
                ),
            ),
            id="specfile_path+synced_files+downstream_package_name",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "spec_source_id": 3,
                "jobs": [get_job_config_dict_build_for_branch()],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        spec_source_id="Source3",
                    ),
                },
                jobs=[
                    get_job_config_build_for_branch(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        spec_source_id="Source3",
                    ),
                ],
            ),
            id="specfile_path+get_job_config_dict_build_for_branch",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "sync_changelog": True,
                "jobs": [get_job_config_dict_simple()],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        sync_changelog=True,
                    ),
                },
                jobs=[
                    get_job_config_simple(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        sync_changelog=True,
                    ),
                ],
            ),
            id="sync_changelog_true",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [get_job_config_dict_simple()],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        sync_changelog=False,
                    ),
                },
                jobs=[
                    get_job_config_simple(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        sync_changelog=False,
                    ),
                ],
            ),
            id="sync_changelog_false_by_default",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "create_sync_note": False,
                "jobs": [get_job_config_dict_simple()],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        create_sync_note=False,
                    ),
                },
                jobs=[
                    get_job_config_simple(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        create_sync_note=False,
                    ),
                ],
            ),
            id="create_sync_note_false",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [get_job_config_dict_simple()],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        create_sync_note=True,
                    ),
                },
                jobs=[
                    get_job_config_simple(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        create_sync_note=True,
                    ),
                ],
            ),
            id="create_sync_note_true_by_default",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "sources": [
                    {
                        "path": "rsync-3.1.3.tar.gz",
                        "url": "https://git.centos.org/sources/rsync/c8s/82e7829",
                    },
                ],
                "jobs": [],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        sources=[
                            SourcesItem(
                                path="rsync-3.1.3.tar.gz",
                                url="https://git.centos.org/sources/rsync/c8s/82e7829",
                            ),
                        ],
                    ),
                },
                jobs=[],
            ),
            id="sources",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "fmf_url": "https://example.com",
                        "fmf_ref": "test_ref",
                        "fmf_path": "test/path",
                    },
                ],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        sync_changelog=False,
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.release,
                        packages={
                            "package": CommonPackageConfig(
                                downstream_package_name="package",
                                specfile_path="fedora/package.spec",
                                sync_changelog=False,
                                fmf_url="https://example.com",
                                fmf_ref="test_ref",
                                fmf_path="test/path",
                            ),
                        },
                    ),
                ],
            ),
            id="sync_changelog_false_by_default",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "tmt_plan": "^packit!",
                        "tf_post_install_script": "echo 'hi packit!'",
                    },
                ],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        sync_changelog=False,
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.release,
                        packages={
                            "package": CommonPackageConfig(
                                downstream_package_name="package",
                                specfile_path="fedora/package.spec",
                                sync_changelog=False,
                                tmt_plan="^packit!",
                                tf_post_install_script="echo 'hi packit!'",
                            ),
                        },
                    ),
                ],
            ),
            id="extra_tf_api_parameters",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "tf_extra_params": {
                            "environments": [
                                {
                                    "pool": "new-pool",
                                    "tmt": {
                                        "context": {
                                            "how": "full",
                                        },
                                    },
                                },
                            ],
                            "test": {"fmf": {"name": "foo"}},
                        },
                    },
                ],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        sync_changelog=False,
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.release,
                        packages={
                            "package": CommonPackageConfig(
                                downstream_package_name="package",
                                specfile_path="fedora/package.spec",
                                sync_changelog=False,
                                tf_extra_params={
                                    "environments": [
                                        {
                                            "pool": "new-pool",
                                            "tmt": {"context": {"how": "full"}},
                                        },
                                    ],
                                    "test": {"fmf": {"name": "foo"}},
                                },
                            ),
                        },
                    ),
                ],
            ),
            id="extra_tf_api_parameters_freeform",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "vm_image_build",
                        "trigger": "pull_request",
                        "image_distribution": "rhel-90",
                        "image_request": {
                            "architecture": "x86_64",
                            "image_type": "aws",
                            "upload_request": {
                                "options": {"share_with_accounts": ["123456789012"]},
                                "type": "aws",
                            },
                        },
                        "image_customizations": {
                            "packages": ["peddle", "board"],
                        },
                    },
                ],
            },
            PackageConfig(
                # parsing makes it inherit: in here we need to be explicit
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.vm_image_build,
                        trigger=JobConfigTriggerType.pull_request,
                        packages={
                            "package": CommonPackageConfig(
                                downstream_package_name="package",
                                specfile_path="fedora/package.spec",
                                image_distribution="rhel-90",
                                image_request={
                                    "architecture": "x86_64",
                                    "image_type": "aws",
                                    "upload_request": {
                                        "options": {
                                            "share_with_accounts": ["123456789012"],
                                        },
                                        "type": "aws",
                                    },
                                },
                                image_customizations={
                                    "packages": ["peddle", "board"],
                                },
                            ),
                        },
                    ),
                ],
            ),
            id="vm-image-build",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "koji_build",
                        "trigger": "commit | koji_build",
                        "sidetag_group": "test",
                        "dependencies": ["some-package"],
                        "dist_git_branches": ["fedora-all"],
                    },
                ],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.koji_build,
                        trigger=JobConfigTriggerType.commit,
                        sidetag_group="test",
                        dependencies=["some-package"],
                        packages={
                            "package": CommonPackageConfig(
                                downstream_package_name="package",
                                specfile_path="fedora/package.spec",
                                dist_git_branches=["fedora-all"],
                            ),
                        },
                    ),
                    JobConfig(
                        type=JobType.koji_build,
                        trigger=JobConfigTriggerType.koji_build,
                        sidetag_group="test",
                        dependencies=["some-package"],
                        packages={
                            "package": CommonPackageConfig(
                                downstream_package_name="package",
                                specfile_path="fedora/package.spec",
                                dist_git_branches=["fedora-all"],
                            ),
                        },
                    ),
                ],
            ),
            id="koji_build_with_multiple_triggers",
        ),
    ],
)
def test_package_config_parse(raw, expected):
    package_config = PackageConfig.get_from_dict(raw_dict=raw)
    assert package_config
    # tests for https://github.com/packit/packit-service/pull/342
    if expected.jobs:
        for j in package_config.jobs:
            assert j.type
    assert package_config == expected


@pytest.mark.parametrize(
    "raw,expected",
    [
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "specfile_path": "somewhere/package.spec",
                    },
                ],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.release,
                        packages={
                            "package": CommonPackageConfig(
                                downstream_package_name="package",
                                specfile_path="somewhere/package.spec",
                            ),
                        },
                    ),
                ],
            ),
            id="override-specfile_path",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "files_to_sync": ["x", "y"],
                "actions": {"post-upstream-clone": "ls"},
                "spec_source_id": "Source0",
                "targets": ["fedora-36"],
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "release",
                        "specfile_path": "somewhere/package.spec",
                        "files_to_sync": ["a", "b", "c"],
                        "actions": {"create-archive": "ls"},
                        "spec_source_id": "Source1",
                    },
                ],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        downstream_package_name="package",
                        specfile_path="fedora/package.spec",
                        files_to_sync=[SyncFilesItem([x], x) for x in ("x", "y")],
                        actions={ActionName.post_upstream_clone: "ls"},
                        _targets=["fedora-36"],
                    ),
                },
                jobs=[
                    JobConfig(
                        type=JobType.copr_build,
                        trigger=JobConfigTriggerType.release,
                        packages={
                            "package": CommonPackageConfig(
                                downstream_package_name="package",
                                specfile_path="somewhere/package.spec",
                                files_to_sync=[
                                    SyncFilesItem([x], x) for x in ("a", "b", "c")
                                ],
                                actions={ActionName.create_archive: "ls"},
                                _targets=["fedora-36"],
                                spec_source_id="Source1",
                            ),
                        },
                    ),
                ],
            ),
            id="override-alot",
        ),
    ],
)
def test_package_config_overrides(raw, expected):
    package_config = PackageConfig.get_from_dict(raw_dict=raw)
    assert package_config == expected


@pytest.mark.parametrize(
    "raw,err_message",
    [
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [{"job": "build", "trigger": "release", "actions": ["a"]}],
            },
            "'dict' required, got <class 'list'>.",
            id="bad_actions",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "fedora/package.spec",
                "jobs": [{"job": "build", "trigger": "release", "synced_files": "a"}],
            },
            "ValidationError",
            id="bad_synced_files",
        ),
    ],
)
def test_package_config_overrides_bad(raw, err_message):
    with pytest.raises(ValidationError) as ex:
        PackageConfig.get_from_dict(raw_dict=raw)
    assert err_message in str(ex)


@pytest.mark.parametrize(
    "raw,expected",
    [
        (
            {
                "specfile_path": "fedora/package.spec",
                "synced_files": ["fedora/package.spec"],
                "jobs": [],
            },
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="fedora/package.spec",
                        synced_files=[
                            SyncFilesItem(
                                src=["fedora/package.spec"],
                                dest="fedora/package.spec",
                            ),
                        ],
                        downstream_package_name="package",
                        upstream_package_name="package",
                    ),
                },
            ),
        ),
    ],
)
def test_package_config_upstream_and_downstream_package_names(raw, expected):
    package_config = PackageConfig.get_from_dict(raw_dict=raw, repo_name="package")
    assert package_config
    assert package_config == expected


def test_dist_git_package_url():
    di = {
        "dist_git_base_url": "https://packit.dev/",
        "downstream_package_name": "packit",
        "dist_git_namespace": "awesome",
        "synced_files": ["fedora/foobar.spec"],
        "specfile_path": "fedora/package.spec",
        "create_pr": False,
    }
    new_pc = PackageConfig.get_from_dict(di)
    pc = PackageConfig(
        packages={
            "packit": CommonPackageConfig(
                dist_git_base_url="https://packit.dev/",
                downstream_package_name="packit",
                dist_git_namespace="awesome",
                synced_files=[
                    SyncFilesItem(
                        src=["fedora/foobar.spec"],
                        dest="fedora/foobar.spec",
                    ),
                ],
                specfile_path="fedora/package.spec",
                create_pr=False,
            ),
        },
        jobs=get_default_job_config(
            dist_git_base_url="https://packit.dev/",
            downstream_package_name="packit",
            dist_git_namespace="awesome",
            synced_files=[
                SyncFilesItem(src=["fedora/foobar.spec"], dest="fedora/foobar.spec"),
            ],
            specfile_path="fedora/package.spec",
            create_pr=False,
        ),
    )
    assert new_pc.specfile_path.endswith("fedora/package.spec")
    assert pc.specfile_path.endswith("fedora/package.spec")
    assert pc == new_pc
    assert pc.dist_git_package_url == "https://packit.dev/awesome/packit.git"
    assert new_pc.dist_git_package_url == "https://packit.dev/awesome/packit.git"
    assert not pc.create_pr


@pytest.mark.parametrize(
    "files,config_name,content,project,spec_path",
    [
        (
            [".packit.yaml", "tox.ini", "setup.py", "packit.spec"],
            ".packit.yaml",
            "files_to_sync:\n"
            "  - packit.spec\n"
            "  - src: {config_name}\n"
            "    dest: .packit2.yaml",
            flexmock(repo="packit"),
            "packit.spec",
        ),
        (
            ["packit.yml", "tox.ini", "setup.py", "packit.spec"],
            "packit.yml",
            "files_to_sync:\n"
            "  - packit.spec\n"
            "  - src: {config_name}\n"
            "    dest: .packit2.yaml",
            flexmock(repo="packit"),
            "packit.spec",
        ),
        (
            ["setup.cfg", "setup.py", "packit.spec", ".zuul.yaml"],
            None,
            "",
            flexmock(repo="packit"),
            "packit.spec",
        ),
    ],
)
def test_get_package_config_from_repo(
    files,
    config_name,
    content,
    project,
    spec_path: Optional[str],
):
    if config_name:
        project.should_receive("get_files").and_return([spec_path]).once()
        project.should_receive("get_file_content").with_args(
            path=config_name,
            ref=None,
        ).and_return(content.format(config_name=config_name)).once()
    project.should_receive("full_repo_name").and_return("a/b")
    project.should_receive("get_files").with_args(ref=None, recursive=False).and_return(
        files,
    )
    config = get_package_config_from_repo(project=project)
    if config_name:
        assert isinstance(config, PackageConfig)
        assert config.specfile_path == spec_path
        assert config.files_to_sync == [
            SyncFilesItem(src=["packit.spec"], dest="packit.spec"),
            SyncFilesItem(src=[config_name], dest=".packit2.yaml"),
        ]
        assert config.create_pr
    else:
        assert config is None


def test_get_package_config_from_repo_explicit_path():
    """Test getting the package config when the path is explicitly specified"""
    project = flexmock(repo="packit")
    config_name = ".distro/source-git.yaml"

    project.should_receive("get_file_content").with_args(
        path=config_name,
        ref=None,
    ).and_return(
        """\
upstream_project_url: https://github.com/packit/packit
downstream_package_name: packit
specfile_path: ".distro/packit.spec"
files_to_sync:
- src: .distro/
  dest: .
  delete: true
  filters:
  - "protect .git*"
  - "protect sources"
  - "exclude source-git.yaml"
  - "exclude .gitignore"
""",
    ).once()
    project.should_receive("full_repo_name").and_return("packit/packit")
    config = get_package_config_from_repo(
        project=project,
        package_config_path=config_name,
    )
    assert config is not None
    assert config.downstream_package_name == "packit"
    assert config.upstream_project_url == "https://github.com/packit/packit"
    assert config.specfile_path == ".distro/packit.spec"


def test_get_package_config_from_repo_empty_no_spec():
    gp = flexmock(GitProject)
    gp.should_receive("get_files").and_return([])
    gp.should_receive("full_repo_name").and_return("a/b")
    gp.should_receive("get_file_content").and_return("# this is a comment")
    gp.should_receive("get_files").with_args(ref=None, recursive=False).and_return(
        ["packit.yaml", "tox.ini"],
    )
    git_project = GitProject(repo="packit", service=GitService(), namespace="")
    with pytest.raises(
        PackitConfigException,
        match="specfile_path' is not specified or no specfile was found in the repo",
    ):
        get_package_config_from_repo(project=git_project)


@pytest.mark.parametrize(
    "content, specfile_path",
    [
        ("{}", "packit.spec"),
        ("{jobs: [{job: build, trigger: commit}]}", "packit.spec"),
        (
            "{downstream_package_name: horkyze, jobs: [{job: build, trigger: commit}]}",
            "horkyze.spec",
        ),
        (
            "{upstream_package_name: slize, jobs: [{job: build, trigger: commit}]}",
            "packit.spec",
        ),
    ],
)
def test_get_package_config_from_repo_spec_file_not_defined(content, specfile_path):
    gp = flexmock(GitProject)
    gp.should_receive("get_files").and_return([specfile_path])
    gp.should_receive("full_repo_name").and_return("a/b")
    gp.should_receive("get_file_content").and_return(content)
    gp.should_receive("get_files").with_args(ref=None, recursive=False).and_return(
        ["packit.yaml", "tox.ini"],
    )
    git_project = GitProject(repo="packit", service=GitService(), namespace="")
    config = get_package_config_from_repo(project=git_project)
    assert isinstance(config, PackageConfig)
    assert config.specfile_path == specfile_path
    assert config.create_pr
    for j in config.jobs:
        assert j.specfile_path == specfile_path
        assert j.downstream_package_name == config.downstream_package_name
        assert j.upstream_package_name == config.upstream_package_name


@pytest.mark.parametrize(
    "package_config, all_synced_files",
    [
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        config_file_path="packit.yaml",
                        specfile_path="file.spec",
                        synced_files=[
                            SyncFilesItem(src=["file.spec"], dest="file.spec"),
                        ],
                    ),
                },
            ),
            [
                SyncFilesItem(src=["file.spec"], dest="file.spec"),
                SyncFilesItem(src=["packit.yaml"], dest="packit.yaml"),
            ],
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        config_file_path="packit.yaml",
                        specfile_path="file.spec",
                        downstream_package_name="package",
                        synced_files=[
                            SyncFilesItem(src=["file.spec"], dest="package.spec"),
                        ],
                    ),
                },
            ),
            [
                SyncFilesItem(src=["file.spec"], dest="package.spec"),
                SyncFilesItem(src=["packit.yaml"], dest="packit.yaml"),
            ],
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        config_file_path="packit.yaml",
                        specfile_path="file.spec",
                        downstream_package_name="package",
                        synced_files=[SyncFilesItem(src=["file.txt"], dest="file.txt")],
                    ),
                },
            ),
            [
                SyncFilesItem(src=["file.txt"], dest="file.txt"),
                SyncFilesItem(src=["file.spec"], dest="package.spec"),
                SyncFilesItem(src=["packit.yaml"], dest="packit.yaml"),
            ],
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        config_file_path="packit.yaml",
                        specfile_path="file.spec",
                        synced_files=[],
                        downstream_package_name="package",
                    ),
                },
            ),
            [
                SyncFilesItem(src=["file.spec"], dest="package.spec"),
                SyncFilesItem(src=["packit.yaml"], dest="packit.yaml"),
            ],
        ),
        (
            PackageConfig(
                packages={
                    "package": CommonPackageConfig(
                        config_file_path="packit.yaml",
                        specfile_path="file.spec",
                        synced_files=[SyncFilesItem(src=["file.txt"], dest="file.txt")],
                        files_to_sync=[
                            SyncFilesItem(src=["file.spec"], dest="file.spec"),
                        ],
                    ),
                },
            ),
            [
                SyncFilesItem(src=["file.spec"], dest="file.spec"),
            ],
        ),
    ],
)
def test_get_all_files_to_sync(package_config, all_synced_files):
    assert package_config.get_all_files_to_sync() == all_synced_files


def test_notifications_section():
    pc = PackageConfig.get_from_dict(
        {"specfile_path": "package.spec"},
        repo_name="package",
    )
    assert not pc.notifications.pull_request.successful_build
    assert pc.notifications.failure_issue.create
    assert pc.notifications.failure_comment.message is None


def test_notifications_section_failure_comment_message():
    message = "my message"
    pc = PackageConfig.get_from_dict(
        {
            "specfile_path": "package.spec",
            "notifications": {"failure_comment": {"message": message}},
        },
        repo_name="package",
    )
    assert not pc.notifications.pull_request.successful_build
    assert pc.notifications.failure_issue.create
    assert pc.notifications.failure_comment.message == message


def test_notifications_section_failure_issue_create_false():
    pc = PackageConfig.get_from_dict(
        {
            "specfile_path": "package.spec",
            "notifications": {"failure_issue": {"create": False}},
        },
        repo_name="package",
    )
    assert not pc.notifications.pull_request.successful_build
    assert not pc.notifications.failure_issue.create


def test_test_command_labels():
    labels = ["label1", "label2"]
    pc = PackageConfig.get_from_dict(
        {
            "specfile_path": "package.spec",
            "test_command": {"default_labels": labels},
        },
        repo_name="package",
    )
    assert pc.test_command.default_labels == labels
    assert not pc.test_command.default_identifier


@pytest.mark.parametrize(
    "package_config_dict, present, absent",
    [
        (
            {
                "specfile_path": "package.spec",
            },
            [],
            [],
        ),
        (
            {"specfile_path": "package.spec", "require": {}},
            [],
            [],
        ),
        (
            {"specfile_path": "package.spec", "require": {"label": {}}},
            [],
            [],
        ),
        (
            {
                "specfile_path": "package.spec",
                "require": {
                    "label": {"present": ["label-1"]},
                },
            },
            ["label-1"],
            [],
        ),
        (
            {
                "specfile_path": "package.spec",
                "require": {
                    "label": {"absent": ["label-1"]},
                },
            },
            [],
            ["label-1"],
        ),
        (
            {
                "specfile_path": "package.spec",
                "require": {
                    "label": {"present": ["label-1"], "absent": ["label-2", "label-3"]},
                },
            },
            ["label-1"],
            ["label-2", "label-3"],
        ),
    ],
)
def test_require(package_config_dict, present, absent):
    pc = PackageConfig.get_from_dict(
        package_config_dict,
        repo_name="package",
    )
    assert pc.require.label.present == present
    assert pc.require.label.absent == absent


def test_test_command_identifiers():
    identifier = "id1"
    pc = PackageConfig.get_from_dict(
        {
            "specfile_path": "package.spec",
            "test_command": {"default_identifier": identifier},
        },
        repo_name="package",
    )
    assert not pc.test_command.default_labels
    assert pc.test_command.default_identifier == identifier


def test_get_raw_dict_with_defaults():
    pc = PackageConfig.get_from_dict(
        {
            "specfile_path": "package.spec",
        },
        repo_name="package",
        config_file_path="path",
    )
    assert pc.get_raw_dict_with_defaults() == {
        "specfile_path": "package.spec",
        "config_file_path": "path",
        "upstream_package_name": "package",
        "downstream_package_name": "package",
    }


def test_get_local_specfile_path():
    assert str(get_local_specfile_path(UP_OSBUILD)) == "osbuild.spec"
    assert not get_local_specfile_path(SYNC_FILES)


@pytest.mark.parametrize(
    "directory, local_first,local_last,config_file_name,res_pc_path",
    [
        ([], False, True, None, Path.cwd() / next(iter(CONFIG_FILE_NAMES))),
        ([], False, False, "different_conf.yaml", "different_conf.yaml"),
    ],
)
def test_get_local_package_config_path(
    directory,
    local_first,
    local_last,
    config_file_name,
    res_pc_path,
):
    flexmock(PosixPath).should_receive("is_file").and_return(True)

    (
        flexmock(packit.config.package_config)
        .should_receive("load_packit_yaml")
        .with_args(Path(res_pc_path))
        .and_return(None)
    )

    (
        flexmock(packit.config.package_config)
        .should_receive("parse_loaded_config")
        .and_return(None)
    )

    get_local_package_config(
        try_local_dir_last=local_last,
        package_config_path=config_file_name,
    )


def test_specfile_path_from_downstream_package_name():
    """
    If 'downstream_package_name' is set in the package configuration,
    'specfile_path' is set to '{downstream_package_name}.spec'
    """
    (
        flexmock(packit.config.package_config)
        .should_receive("load_packit_yaml")
        .and_return({"downstream_package_name": "cockpit"})
    )
    assert (
        get_local_package_config(
            package_config_path=".packit.yaml",
            repo_name="cockpit",
        ).specfile_path
        == PackageConfig(
            packages={"cockpit": CommonPackageConfig(specfile_path="cockpit.spec")},
        ).specfile_path
    )


@pytest.mark.parametrize(
    "package_config",
    [
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    downstream_package_name="package",
                    upstream_package_name="package",
                    synced_files=[
                        SyncFilesItem(
                            src=["fedora/package.spec"],
                            dest="fedora/package.spec",
                        ),
                    ],
                ),
            },
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    downstream_package_name="package",
                    upstream_package_name="package",
                    synced_files=[
                        SyncFilesItem(
                            src=["fedora/package.spec"],
                            dest="fedora/package.spec",
                        ),
                    ],
                    files_to_sync=[
                        SyncFilesItem(
                            src=["fedora/package.spec"],
                            dest="fedora/package.spec",
                        ),
                    ],
                ),
            },
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    downstream_package_name="package",
                    upstream_package_name="package",
                    synced_files=[
                        SyncFilesItem(src=["fedora/p.spec"], dest="fedora/p.spec"),
                    ],
                    files_to_sync=[
                        SyncFilesItem(
                            src=["fedora/package.spec"],
                            dest="fedora/package.spec",
                        ),
                    ],
                ),
            },
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    downstream_package_name="package",
                    upstream_package_name="package",
                    files_to_sync=[
                        SyncFilesItem(
                            src=["fedora/package.spec"],
                            dest="fedora/package.spec",
                        ),
                    ],
                ),
            },
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    downstream_package_name="package",
                    synced_files=[
                        SyncFilesItem(
                            src=["fedora/package.spec"],
                            dest="fedora/package.spec",
                        ),
                    ],
                ),
            },
            jobs=[
                get_job_config_full(
                    specfile_path="fedora/package.spec",
                    downstream_package_name="package",
                    synced_files=[
                        SyncFilesItem(
                            src=["fedora/package.spec"],
                            dest="fedora/package.spec",
                        ),
                    ],
                ),
            ],
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    actions={
                        ActionName.pre_sync: "some/pre-sync/command --option",
                        ActionName.get_current_version: "get-me-version",
                    },
                    upstream_project_url="https://github.com/asd/qwe",
                    upstream_package_name="qwe",
                    dist_git_base_url="https://something.wicked",
                    downstream_package_name="package",
                    spec_source_id="Source1",
                ),
            },
            jobs=[],
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    actions={
                        ActionName.pre_sync: "some/pre-sync/command --option",
                        ActionName.get_current_version: "get-me-version",
                    },
                    upstream_project_url="https://github.com/asd/qwe",
                    upstream_package_name="qwe",
                    dist_git_base_url="https://something.wicked",
                    downstream_package_name="package",
                    spec_source_id="Source1",
                    sources=[
                        SourcesItem(
                            path="rsync-3.1.3.tar.gz",
                            url="https://git.centos.org/sources/rsync/c8s/82e7829",
                        ),
                    ],
                ),
            },
            jobs=[],
        ),
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    actions={
                        ActionName.pre_sync: "some/pre-sync/command --option",
                        ActionName.get_current_version: "get-me-version",
                    },
                    upstream_project_url="https://github.com/asd/qwe",
                    upstream_package_name="qwe",
                    srpm_build_deps=["make", "tar", "findutils"],
                ),
            },
            jobs=[],
        ),
        PackageConfig(
            packages={
                "main": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    upstream_project_url="https://github.com/asd/qwe",
                    upstream_package_name="qwe",
                    srpm_build_deps=[],
                    status_name_template="packit-job",
                ),
            },
            jobs=[],
        ),
    ],
)
def test_serialize_and_deserialize(package_config):
    schema = PackageConfigSchema()
    serialized = schema.dump(package_config)
    new_package_config = schema.load(serialized)
    assert package_config == new_package_config


@pytest.mark.parametrize(
    "package_config",
    [
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    specfile_path="fedora/package.spec",
                    config_file_path=".packit.yaml",
                    downstream_package_name="package",
                    upstream_package_name="package",
                ),
            },
        ),
    ],
)
def test_files_to_sync_after_dump(package_config):
    schema = PackageConfigSchema()
    assert len(package_config.get_all_files_to_sync()) == 2
    serialized = schema.dump(package_config)
    new_package_config = schema.load(serialized)
    assert len(new_package_config.get_all_files_to_sync()) == 2


def test_get_specfile_sync_files_item():
    pc = PackageConfig(
        packages={
            "ogr": CommonPackageConfig(
                specfile_path="fedora/python-ogr.spec",
                downstream_package_name="python-ogr",
            ),
        },
    )
    upstream_specfile_path = "fedora/python-ogr.spec"
    downstream_specfile_path = "python-ogr.spec"

    assert pc.get_specfile_sync_files_item() == SyncFilesItem(
        src=[upstream_specfile_path],
        dest=downstream_specfile_path,
    )
    assert pc.get_specfile_sync_files_item(from_downstream=True) == SyncFilesItem(
        src=[downstream_specfile_path],
        dest=upstream_specfile_path,
    )


def test_get_specfile_sync_files_nodownstreamname_item():
    pc = PackageConfig(
        packages={
            "package": CommonPackageConfig(specfile_path="fedora/python-ogr.spec"),
        },
    )
    upstream_specfile_path = "fedora/python-ogr.spec"
    downstream_specfile_path = "python-ogr.spec"

    assert pc.get_specfile_sync_files_item() == SyncFilesItem(
        src=[upstream_specfile_path],
        dest=downstream_specfile_path,
    )
    assert pc.get_specfile_sync_files_item(from_downstream=True) == SyncFilesItem(
        src=[downstream_specfile_path],
        dest=upstream_specfile_path,
    )


@pytest.mark.parametrize(
    "raw",
    [
        {
            "packages": {"package": {}},
            "jobs": [
                {
                    "job": "copr_build",
                    "trigger": "release",
                    "targets": ["fedora-stable", "fedora-development"],
                },
            ],
        },
        {
            "packages": {"package": {}},
            "jobs": [
                {
                    "job": "tests",
                    "trigger": "release",
                    "targets": ["fedora-stable", "fedora-development"],
                },
            ],
        },
    ],
)
def test_package_config_specfile_not_present_raise(raw):
    with pytest.raises(ValidationError, match="'specfile_path' is not specified"):
        PackageConfig.get_from_dict(raw_dict=raw)


@pytest.mark.parametrize(
    "raw",
    [
        {
            "jobs": [
                {
                    "job": "tests",
                    "trigger": "commit",
                    "targets": ["fedora-stable", "fedora-development"],
                    "skip_build": True,
                    "manual_trigger": True,
                },
            ],
        },
        {
            "jobs": [
                {
                    "job": "tests",
                    "trigger": "commit",
                    "targets": ["fedora-stable", "fedora-development"],
                    "skip_build": True,
                    "manual_trigger": True,
                },
                {
                    "job": "tests",
                    "trigger": "pull_request",
                    "metadata": {
                        "targets": ["fedora-stable", "fedora-development"],
                        "skip_build": True,
                    },
                },
            ],
        },
    ],
)
def test_specfile_path_not_defined_in_test_only_jobs(raw):
    """Packages in test jobs, which are configured to not to require building
    a package don't need a 'specfile_path' set.

    Note: don't set 'downstream_package_name' either, as that implicitly sets
    'specfile_path' :-)
    """
    assert PackageConfig.get_from_dict(raw_dict=raw, repo_name="package")


@pytest.mark.parametrize(
    "package_name,result",
    ((None, None), ("baz", "https://foo/bar/baz.git")),
)
def test_pc_dist_git_package_url_has_no_None(package_name, result):
    assert (
        PackageConfig(
            packages={
                "package": CommonPackageConfig(
                    downstream_package_name=package_name,
                    dist_git_base_url="https://foo/",
                    dist_git_namespace="bar",
                ),
            },
        ).dist_git_package_url
        == result
    )


def test_deepcopy():
    """Make sure the config object can be deepcopied"""
    d = {"package_config": PackageConfig(packages={"package": CommonPackageConfig()})}
    copy.deepcopy(d)


def test_load_is_not_destructive():
    """Check that the data loaded by PackageConfigSchema is not modified,
    and stays as it was passed for loading.
    """
    data = {"downstream_package_name": "package", "specfile_path": "package.spec"}
    original = copy.deepcopy(data)
    PackageConfigSchema().load(data)
    assert data == original


def test_handle_metadata():
    """Check that the deprecated 'metadata' field in 'jobs' is handled
    correctly.
    """
    data = {
        "downstream_package_name": "package",
        "specfile_path": "package.spec",
        "jobs": [
            {
                "job": "copr_build",
                "trigger": "pull_request",
                "metadata": {
                    "dist-git-branch": "shrek",
                },
            },
        ],
    }
    pc = PackageConfigSchema().load(data)
    assert pc.jobs[0].dist_git_branches == ["shrek"]


@pytest.mark.parametrize(
    "config",
    [
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "package.spec",
                "upstream_project_name": "upstream_package",
            },
            id="top_level",
        ),
    ],
)
def test_deprecated_keys_renamed(config):
    """Deprecated keys are renamed when defined on the top-level."""
    pc = PackageConfigSchema().load(config)
    assert pc.upstream_package_name == "upstream_package"
    assert pc.packages["package"].upstream_package_name == "upstream_package"


@pytest.mark.parametrize(
    "config",
    [
        pytest.param(
            {
                "packages": {
                    "package": {
                        "specfile_path": "package.spec",
                        "upstream_project_name": "upstream_package",
                    },
                },
            },
            id="in_packages",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "packages.spec",
                "upstream_package_name": "package",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "upstream_project_name": "upstream_package",
                    },
                ],
            },
            id="in_job",
        ),
        pytest.param(
            {
                "downstream_package_name": "package",
                "specfile_path": "packages.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "packages": {
                            "package": {
                                "upstream_project_name": "upstream_package",
                            },
                        },
                    },
                ],
            },
            id="in_job_packages",
        ),
    ],
)
def test_deprecated_keys_invalid(config):
    """Deprecated keys are not renamed in job and package objects and trigger a
    ValidationError."""
    with pytest.raises(ValidationError, match=r"upstream_project_name.+Unknown field"):
        PackageConfigSchema().load(config)


def test_loading_packageless_config():
    """Check that configuration data without any 'packages' key is correctly
    loaded."""
    config_data = {
        "downstream_package_name": "foo",
        "upstream_package_name": "foo",
        "issue_repository": "https://github.com/foo/bar",
        "actions": {
            "get-current-version": [
                "python3 setup.py --version",
            ],
        },
        "files_to_sync": ["packaging/foo.spec", "packit.yaml"],
        "srpm_build_deps": ["git", "python"],
        "specfile_path": "packaging/foo.spec",
        "jobs": [
            {
                "job": "copr_build",
                "trigger": "pull_request",
                "targets": ["fedora-stable", "epel-9"],
            },
            {
                "job": "copr_build",
                "trigger": "commit",
                "targets": ["fedora-stable", "epel-9"],
                "project": "foo-dev",
                "list_on_homepage": True,
                "preserve_project": True,
            },
        ],
    }
    loaded_config = PackageConfigSchema().load(config_data)

    # Prepare objects in the expected configuration
    top_level_config = {k: v for k, v in config_data.items() if k != "jobs"}
    top_level_config["files_to_sync"] = [
        SyncFilesItem(src=[item], dest=item)
        for item in top_level_config["files_to_sync"]
    ]
    top_level_config["actions"] = {
        ActionName(k): v for k, v in top_level_config["actions"].items()
    }
    job_config = [
        {
            f"_{k}" if k == "targets" else k: v
            for k, v in job.items()
            if k not in {"job", "trigger"}
        }
        for job in config_data["jobs"]
    ]
    expected_config = PackageConfig(
        packages={"foo": CommonPackageConfig(paths=["./"], **top_level_config)},
        jobs=[
            JobConfig(
                type=JobType.copr_build,
                trigger=JobConfigTriggerType.pull_request,
                packages={
                    "foo": CommonPackageConfig(
                        paths=["./"],
                        **(top_level_config | job_config[0]),
                    ),
                },
            ),
            JobConfig(
                type=JobType.copr_build,
                trigger=JobConfigTriggerType.commit,
                packages={
                    "foo": CommonPackageConfig(
                        paths=["./"],
                        **(top_level_config | job_config[1]),
                    ),
                },
            ),
        ],
    )
    assert loaded_config == expected_config


def test_package_naming():
    """Either the key from 'packages' or 'downstream_package_name' is used to name
    a package."""
    config1 = PackageConfigSchema().load(
        {
            "downstream_package_name": "baar",
            "specfile_path": "fedora/baar.spec",
        },
    )
    config2 = PackageConfigSchema().load(
        {
            "packages": {
                "baar": {
                    "specfile_path": "fedora/baar.spec",
                },
            },
        },
    )
    assert config1 == config2
    assert config1.downstream_package_name == config2.downstream_package_name == "baar"
    # check the default value for 'paths'
    assert config1.paths == config2.paths == ["./"]
    assert config1.packages["baar"].paths == config2.packages["baar"].paths == ["./"]


def test_multiple_packages():
    """A configuration with multiple packages can be loaded.
    Values can be accessed only through the 'packages' attribute."""
    config = {
        "packages": {
            "foo": {
                "specfile_path": "foo/foo.spec",
                "paths": ["foo"],
            },
            "baar": {
                "specfile_path": "baar/baar.spec",
                "paths": ["baar"],
            },
            "jeee": {
                "specfile_path": "jeee/jeee.spec",
                "paths": ["jeee"],
            },
        },
    }
    pc = PackageConfigSchema().load(config)
    assert pc.packages.keys() == pc.jobs[0].packages.keys() == config["packages"].keys()
    # Attributes not related to packages are accessible
    pc.packages = pc.packages
    pc.jobs = pc.jobs
    pc.jobs[0].type = pc.jobs[0].type
    pc.jobs[0].trigger = pc.jobs[0].trigger
    # Accessing package related attributes without 'packages' raises an error
    with pytest.raises(AttributeError, match="ambiguous to get"):
        _ = pc.jobs[0].enable_net
    with pytest.raises(AttributeError, match="ambiguous to get"):
        _ = pc.enable_net
    with pytest.raises(AttributeError, match="ambiguous to set"):
        pc.jobs[0].enable_net = False
    with pytest.raises(AttributeError, match="ambiguous to set"):
        pc.enable_net = False


@pytest.mark.parametrize(
    "data",
    [
        pytest.param(
            {
                "downstream_package_name": "apple",
                "specfile_path": "apple.spec",
                "jobs": [
                    {"job": "copr_build", "trigger": "pull_request"},
                    {"job": "tests", "trigger": "pull_request"},
                ],
            },
            id="no_package_selected",
        ),
        pytest.param(
            {
                "downstream_package_name": "apple",
                "specfile_path": "apple.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "packages": ["apple"],
                    },
                    {
                        "job": "tests",
                        "trigger": "pull_request",
                        "packages": {"apple": {}},
                    },
                ],
            },
            id="package_selected",
        ),
        pytest.param(
            {
                "downstream_package_name": "apple",
                "specfile_path": "apple.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "packages": {"apple": {}},
                    },
                    {"job": "tests", "trigger": "pull_request", "packages": ["apple"]},
                ],
            },
            id="package_selected",
        ),
    ],
)
def test_selecting_packages_in_jobs(data):
    """Jobs can select to work with all or just a subset of the packages defined top-level"""
    expected_config = PackageConfig(
        packages={
            "apple": CommonPackageConfig(
                downstream_package_name="apple",
                specfile_path="apple.spec",
            ),
        },
        jobs=[
            JobConfig(
                type=JobType.copr_build,
                trigger=JobConfigTriggerType.pull_request,
                packages={
                    "apple": CommonPackageConfig(
                        downstream_package_name="apple",
                        specfile_path="apple.spec",
                    ),
                },
            ),
            JobConfig(
                type=JobType.tests,
                trigger=JobConfigTriggerType.pull_request,
                packages={
                    "apple": CommonPackageConfig(
                        downstream_package_name="apple",
                        specfile_path="apple.spec",
                    ),
                },
            ),
        ],
    )
    loaded_config = PackageConfigSchema().load(data)
    assert loaded_config == expected_config


@pytest.mark.parametrize(
    "data, error",
    [
        pytest.param(
            {
                "downstream_package_name": "apple",
                "specfile_path": "apple.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "packages": "apple",
                    },
                ],
            },
            r"'str'.+instead of 'list' or 'dict'",
            id="package_is_str",
        ),
        pytest.param(
            {
                "downstream_package_name": "apple",
                "specfile_path": "apple.spec",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "packages": ["pear"],
                    },
                ],
            },
            r"Undefined.+: pear\.",
            id="package_is_not_present",
        ),
    ],
)
def test_package_error_in_job(data, error):
    """The 'packages' key in a job is of a wrong type, or references a
    packages which is not defined on the top-level 'packages' dict.
    """
    with pytest.raises(ValidationError, match=error):
        PackageConfigSchema().load(data)


@pytest.mark.parametrize(
    "data",
    [
        pytest.param(
            {
                "packages": {
                    "apple": {
                        "specfile_path": "apple.spec",
                        "paths": ["apple"],
                    },
                    "pear": {
                        "specfile_path": "pear.spec",
                        "paths": ["pear"],
                    },
                },
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                    },
                    {"job": "tests", "trigger": "pull_request", "packages": ["pear"]},
                ],
            },
            id="none_selected",
        ),
        pytest.param(
            {
                "packages": {
                    "apple": {
                        "specfile_path": "apple.spec",
                        "paths": ["apple"],
                    },
                    "pear": {
                        "specfile_path": "pear.spec",
                        "paths": ["pear"],
                    },
                },
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "packages": ["pear", "apple"],
                    },
                    {"job": "tests", "trigger": "pull_request", "packages": ["pear"]},
                ],
            },
            id="multiple_selected",
        ),
    ],
)
def test_multiple_packages_in_jobs(data):
    """Check that it's possible to select one or more packages in jobs."""
    apple_config = CommonPackageConfig(
        downstream_package_name="apple",
        specfile_path="apple.spec",
        paths=["apple"],
    )
    pear_config = CommonPackageConfig(
        downstream_package_name="pear",
        specfile_path="pear.spec",
        paths=["pear"],
    )
    expected_config = PackageConfig(
        packages={
            "apple": apple_config,
            "pear": pear_config,
        },
        jobs=[
            JobConfig(
                type=JobType.copr_build,
                trigger=JobConfigTriggerType.pull_request,
                packages={
                    "apple": apple_config,
                    "pear": pear_config,
                },
            ),
            JobConfig(
                type=JobType.tests,
                trigger=JobConfigTriggerType.pull_request,
                packages={
                    "pear": pear_config,
                },
            ),
        ],
    )
    loaded_config = PackageConfigSchema().load(data)
    assert loaded_config == expected_config


@pytest.mark.parametrize(
    "data",
    [
        pytest.param(
            {
                "specfile_path": "fedora/foo.spec",
                "packages": {
                    "foo": {},
                },
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "specfile_path": "copr/foo.spec",
                    },
                ],
            },
            id="override_in_job",
        ),
        pytest.param(
            {
                "specfile_path": "fedora/foo.spec",
                "downstream_package_name": "foo",
                "jobs": [
                    {
                        "job": "copr_build",
                        "trigger": "pull_request",
                        "packages": {
                            "foo": {
                                "specfile_path": "copr/foo.spec",
                            },
                        },
                    },
                ],
            },
            id="override_in_job",
        ),
    ],
)
def test_configuring_packages_in_jobs(data):
    """Configuration of individual or all packages can be overwritten in jobs"""
    package_foo_config = CommonPackageConfig(
        downstream_package_name="foo",
        specfile_path="fedora/foo.spec",
    )
    job_foo_config = CommonPackageConfig(
        downstream_package_name="foo",
        specfile_path="copr/foo.spec",
    )
    expected_config = PackageConfig(
        packages={"foo": package_foo_config},
        jobs=[
            JobConfig(
                type=JobType.copr_build,
                trigger=JobConfigTriggerType.pull_request,
                packages={"foo": job_foo_config},
            ),
        ],
    )
    loaded_config = PackageConfigSchema().load(data)
    assert loaded_config == expected_config


def test_find_remote_package_config_no_commit():
    exception = GithubAPIException()
    exception.__cause__ = GithubException(404, None, None)

    project = flexmock()
    project.should_receive("get_files").and_raise(exception)

    assert find_remote_package_config(project, ref=None) is None


def test_find_remote_package_config_should_raise():
    exception = GithubAPIException()
    exception.__cause__ = GithubException(403, None, None)

    project = flexmock()
    project.should_receive("get_files").and_raise(exception)

    with pytest.raises(GithubAPIException):
        find_remote_package_config(project, ref=None)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import textwrap

import pytest
from flexmock import flexmock

from packit.constants import COMMIT_ACTION_DIVIDER
from packit.exceptions import PackitException
from packit.utils.repo import (
    get_commit_hunks,
    get_commit_link,
    get_commit_message_from_action,
    get_message_from_metadata,
    get_metadata_from_message,
    get_namespace_and_repo_name,
    get_tag_link,
    git_patch_ish,
    git_remote_url_to_https_url,
)


@pytest.mark.parametrize(
    "url,namespace,repo_name",
    [
        ("https://github.com/org/name", "org", "name"),
        ("https://github.com/org/name/", "org", "name"),
        ("https://github.com/org/name.git", "org", "name"),
        ("git@github.com:org/name", "org", "name"),
        ("git@github.com:org/name.git", "org", "name"),
    ],
)
def test_get_ns_repo(url, namespace, repo_name):
    assert get_namespace_and_repo_name(url) == (namespace, repo_name)


def test_get_ns_repo_exc():
    url = "git@github.com"
    with pytest.raises(PackitException) as ex:
        get_namespace_and_repo_name(url)
    msg = f"Invalid URL format, can't obtain namespace and repository name: {url}"
    assert msg in str(ex.value)


@pytest.mark.parametrize("inp", ["/", None, ""])
def test_remote_to_https_invalid(inp):
    assert git_remote_url_to_https_url(inp) == ""


@pytest.mark.parametrize(
    "inp",
    [
        "https://github.com/packit/packit",
        "https://github.com/packit/packit.git",
        "http://github.com/packit/packit",
        "http://github.com/packit/packit.git",
        "http://www.github.com/packit/packit",
    ],
)
def test_remote_to_https_unchanged(inp):
    assert git_remote_url_to_https_url(inp) == inp


@pytest.mark.parametrize(
    "inp,with_suffix,result",
    [
        ("git@github.com:packit/ogr", True, "https://github.com/packit/ogr"),
        (
            "ssh://ttomecek@pkgs.fedoraproject.org/rpms/alot.git",
            True,
            "https://pkgs.fedoraproject.org/rpms/alot.git",
        ),
        ("www.github.com/packit/packit", True, "https://www.github.com/packit/packit"),
        ("github.com/packit/packit", True, "https://github.com/packit/packit"),
        ("git://github.com/packit/packit", True, "https://github.com/packit/packit"),
        (
            "git+https://github.com/packit/packit.git",
            True,
            "https://github.com/packit/packit.git",
        ),
        (
            "git+https://github.com/packit/packit.git",
            False,
            "https://github.com/packit/packit",
        ),
        (
            "https://github.com/packit/packit.git",
            False,
            "https://github.com/packit/packit",
        ),
    ],
)
def test_remote_to_https(inp, with_suffix, result):
    assert git_remote_url_to_https_url(inp, with_suffix) == result


@pytest.mark.parametrize(
    "inp,outp",
    [
        pytest.param(
            """
            """,
            """
            """,
            id="empty-patch",
        ),
        # There are some \t characters in the strings bellow,
        # and thats expected.
        pytest.param(
            textwrap.dedent(
                """
                Short description: NSCD must use nscd user.
                Author(s): Fedora glibc team <glibc@lists.fedoraproject.org>
                Origin: PATCH
                Upstream status: not-needed

                Fedora-specific configuration adjustment to introduce the nscd user.
                (Upstream does not assume this user exists.)

                diff -Nrup a/nscd/nscd.conf b/nscd/nscd.conf
                --- a/nscd/nscd.conf	2012-06-05 07:42:49.000000000 -0600
                +++ b/nscd/nscd.conf	2012-06-07 12:15:21.818318670 -0600
                @@ -33,7 +33,7 @@
                 #	logfile			/var/log/nscd.log
                 #	threads			4
                 #	max-threads		32
                -#	server-user		nobody
                +	server-user		nscd
                 #	stat-user		somebody
                    debug-level		0
                 #	reload-count		5
            """,
            ),
            textwrap.dedent(
                """
                Short description: NSCD must use nscd user.
                Author(s): Fedora glibc team <glibc@lists.fedoraproject.org>
                Origin: PATCH
                Upstream status: not-needed

                Fedora-specific configuration adjustment to introduce the nscd user.
                (Upstream does not assume this user exists.)

                diff --git a/nscd/nscd.conf b/nscd/nscd.conf
                --- a/nscd/nscd.conf
                +++ b/nscd/nscd.conf
                @@ -33,7 +33,7 @@
                 #	logfile			/var/log/nscd.log
                 #	threads			4
                 #	max-threads		32
                -#	server-user		nobody
                +	server-user		nscd
                 #	stat-user		somebody
                    debug-level		0
                 #	reload-count		5
            """,
            ),
            id="remove-timestamps",
        ),
        pytest.param(
            textwrap.dedent(
                """
                Short description: NSCD must use nscd user.
                Author(s): Fedora glibc team <glibc@lists.fedoraproject.org>
                Origin: PATCH
                Upstream status: not-needed

                Fedora-specific configuration adjustment to introduce the nscd user.
                (Upstream does not assume this user exists.)

                --- a/nscd/nscd.conf	2012-06-05 07:42:49.000000000 -0600
                +++ b/nscd/nscd.conf	2012-06-07 12:15:21.818318670 -0600
                @@ -33,7 +33,7 @@
                 #	logfile			/var/log/nscd.log
                 #	threads			4
                 #	max-threads		32
                -#	server-user		nobody
                +	server-user		nscd
                 #	stat-user		somebody
                    debug-level		0
                 #	reload-count		5
            """,
            ),
            textwrap.dedent(
                """
                Short description: NSCD must use nscd user.
                Author(s): Fedora glibc team <glibc@lists.fedoraproject.org>
                Origin: PATCH
                Upstream status: not-needed

                Fedora-specific configuration adjustment to introduce the nscd user.
                (Upstream does not assume this user exists.)

                diff --git a/nscd/nscd.conf b/nscd/nscd.conf
                --- a/nscd/nscd.conf
                +++ b/nscd/nscd.conf
                @@ -33,7 +33,7 @@
                 #	logfile			/var/log/nscd.log
                 #	threads			4
                 #	max-threads		32
                -#	server-user		nobody
                +	server-user		nscd
                 #	stat-user		somebody
                    debug-level		0
                 #	reload-count		5
            """,
            ),
            id="add-missing-diff",
        ),
    ],
)
def test_git_patch_ish(inp, outp):
    assert git_patch_ish(inp) == outp


@pytest.mark.parametrize(
    "source, result",
    [
        pytest.param("", None, id="empty message"),
        pytest.param("One sentence", None, id="one sentence"),
        pytest.param("One sentence\n", None, id="one sentence with end-line"),
        pytest.param(
            "One sentence\n\n",
            None,
            id="one sentence with multiple end-lines",
        ),
        pytest.param("One sentence\nSecond sentence\n", None, id="two sentences"),
        pytest.param("One sentence\nSecond sentence\n", None, id="two sentences"),
        pytest.param(
            "key: value",
            {"key": "value"},
            id="one key-value",
        ),
        pytest.param(
            "key: value\n",
            {"key": "value"},
            id="one key-value with empty-line",
        ),
        pytest.param(
            "key: value\nsecond_key: value",
            {"key": "value", "second_key": "value"},
            id="two key-values",
        ),
        pytest.param(
            "One sentence\nkey: value\n",
            {"key": "value"},
            id="one sentence and one key-value with empty-line",
        ),
        pytest.param(
            "Once sentence\nSecond sentence.\nkey: value",
            {"key": "value"},
            id="two sentences and one key-value",
        ),
        pytest.param(
            "Once sentence\n\nSecond sentence.\n\n\nkey: value",
            {"key": "value"},
            id="two sentences and one key-value with few empty lines",
        ),
        pytest.param(
            "Once sentence\nSecond sentence.\nkey: value\nother: key",
            {"key": "value", "other": "key"},
            id="two sentences and one key-value",
        ),
        pytest.param(
            "key: value\nsentence at the end",
            None,
            id="sentence at the end",
        ),
        pytest.param(
            "key: [a,b,c]",
            {"key": ["a", "b", "c"]},
            id="list as a value",
        ),
        pytest.param(
            "key:\n- a\n- b\n- c",
            {"key": ["a", "b", "c"]},
            id="list as a value in separate lines",
        ),
        pytest.param(
            "Sentence: with colon\nanother sentence without colon",
            None,
            id="colon in the sentence",
        ),
        pytest.param(
            "Sentence with colon: in the middle\nanother sentence without colon",
            None,
            id="colon in the sentence in the middle",
        ),
    ],
)
def test_get_metadata_from_message(source, result):
    assert get_metadata_from_message(commit=flexmock(message=source)) == result


@pytest.mark.parametrize(
    "metadata, header, result",
    [
        pytest.param({}, None, "", id="empty dict"),
        pytest.param({"key": "value"}, None, "key: value\n", id="single key-value"),
        pytest.param(
            {"key": "value", "other": "value"},
            None,
            "key: value\nother: value\n",
            id="multiple key-values",
        ),
    ],
)
def test_get_message_from_metadata(metadata, header, result):
    assert get_message_from_metadata(metadata, header) == result


def test_get_commit_hunks_single_change():
    commit = "abcdefgh"
    git_api = flexmock()
    git_api.should_receive("show").with_args(
        commit,
        format="",
        color="never",
    ).and_return(
        """\
diff --git a/test2 b/test2
deleted file mode 100644
index 1c7858b..0000000
--- a/test2
+++ /dev/null
@@ -1 +0,0 @@
-deleted
        """,
    )
    repository = flexmock(git=git_api)
    hunks = get_commit_hunks(repository, commit)
    assert len(hunks) == 1
    assert "diff --git a/test2 b/test2" in hunks[0]
    assert "-deleted" in hunks[0]


def test_get_commit_hunks_multiple_changes():
    commit = "abcdefgh"
    git_api = flexmock()
    git_api.should_receive("show").with_args(
        commit,
        format="",
        color="never",
    ).and_return(
        """\
diff --git a/a b/a
new file mode 100644
index 0000000..7898192
--- /dev/null
+++ b/a
@@ -0,0 +1 @@
+a
diff --git a/b.txt b/b.txt
new file mode 100644
index 0000000..6178079
--- /dev/null
+++ b/b.txt
@@ -0,0 +1 @@
+b
        """,
    )
    repository = flexmock(git=git_api)
    hunks = get_commit_hunks(repository, commit)
    assert len(hunks) == 2
    assert "diff --git a/a b/a" in hunks[0]
    assert "+a" in hunks[0]
    assert "diff --git a/b.txt b/b.txt" in hunks[1]
    assert "+b" in hunks[1]


@pytest.mark.parametrize(
    "action_output",
    (
        pytest.param(
            None,
            id="no action defined",
        ),
        pytest.param(
            [],
            id="no output produced",
        ),
        pytest.param(
            # we keep a newline at the end of stdout, because we assume it's been
            # printed out and follows the convention of finishing with ‹\n›
            [
                "debug output\n",
                f"debug including divider\n{COMMIT_ACTION_DIVIDER}",
            ],
            id="nothing after divider",
        ),
        pytest.param(
            # we keep a newline at the end of stdout, because we assume it's been
            # printed out and follows the convention of finishing with ‹\n›
            [COMMIT_ACTION_DIVIDER],
            id="only divider",
        ),
        pytest.param(
            [COMMIT_ACTION_DIVIDER, "\n\n\n\n\n\ncommit body"],
            id="lot of newlines, no commit title",
        ),
    ),
)
def test_get_commit_message_from_action_default(action_output):
    # This test case contains only outputs that produce defaults
    title, body = get_commit_message_from_action(
        action_output,
        "default title",
        "default description",
    )

    assert title == "default title"
    assert body == "default description"


@pytest.mark.parametrize(
    "action_output, expected_title, expected_body",
    (
        pytest.param(
            ["debug\n", COMMIT_ACTION_DIVIDER, "commit title\n"],
            "commit title",
            "",
            id="only commit title given",
        ),
        pytest.param(
            ["debug\n", COMMIT_ACTION_DIVIDER, "commit title\n\ncommit body\n"],
            "commit title",
            "commit body",
            id="both title and body given",
        ),
        pytest.param(
            [COMMIT_ACTION_DIVIDER, "commit title\n"],
            "commit title",
            "",
            id="only commit title given; no debug messages",
        ),
        pytest.param(
            [COMMIT_ACTION_DIVIDER, "commit title\n\ncommit body\n"],
            "commit title",
            "commit body",
            id="both title and body given; no debug messages",
        ),
        pytest.param(
            ["commit title\n\n", "commit body\n"],
            "commit title",
            "commit body",
            id="both title and body given; no divider and debug output present",
        ),
        pytest.param(
            ["commit title\n"],
            "commit title",
            "",
            id="only commit title given; no divider and debug output present",
        ),
    ),
)
def test_get_commit_message_from_action(action_output, expected_title, expected_body):
    title, body = get_commit_message_from_action(
        action_output,
        "default title",
        "default description",
    )

    assert title == expected_title
    assert body == expected_body


@pytest.mark.parametrize(
    "git_url, commit, result",
    [
        pytest.param(
            "https://github.com/packit/packit-service",
            "abcdefg",
            "https://github.com/packit/packit-service/commit/abcdefg",
        ),
        pytest.param(
            "https://gitlab.com/packit/packit-service",
            "abcdefg",
            "https://gitlab.com/packit/packit-service/-/commit/abcdefg",
        ),
        pytest.param(
            "https://gitlab.gnome.org/packit/packit-service",
            "abcdefg",
            "https://gitlab.gnome.org/packit/packit-service/-/commit/abcdefg",
        ),
        pytest.param(
            "https://pagure.io/packit/packit-service",
            "abcdefg",
            "https://pagure.io/packit/packit-service/c/abcdefg",
        ),
    ],
)
def test_get_commit_link(git_url, commit, result):
    assert get_commit_link(git_url, commit) == result


@pytest.mark.parametrize(
    "git_url, tag, result",
    [
        pytest.param(
            "https://github.com/packit/packit-service",
            "1.0.0",
            "https://github.com/packit/packit-service/releases/tag/1.0.0",
        ),
        pytest.param(
            "https://gitlab.com/packit/packit-service",
            "1.0.0",
            "https://gitlab.com/packit/packit-service/-/tags/1.0.0",
        ),
        pytest.param(
            "https://gitlab.gnome.org/packit/packit-service",
            "1.0.0",
            "https://gitlab.gnome.org/packit/packit-service/-/tags/1.0.0",
        ),
        pytest.param("https://pagure.io/packit/packit-service", "1.0.0", ""),
    ],
)
def test_get_tag_link(git_url, tag, result):
    assert get_tag_link(git_url, tag) == result
